#!/usr/bin/env python3

# Copyright © 2020-2025, Meheret Tesfaye Batu <meherett.batu@gmail.com>
# Distributed under the MIT software license, see the accompanying
# file COPYING or https://opensource.org/license/mit

from typing import (
    Dict, List, Type, Union
)

from ..exceptions import (
    ECCError, PublicKeyError
)
from ..utils import get_bytes
from .kholaw import (
    KholawEd25519ECC, KholawEd25519Point, KholawEd25519PublicKey, KholawEd25519PrivateKey
)
from .slip10 import (
    SLIP10Ed25519ECC, SLIP10Ed25519Point, SLIP10Ed25519PublicKey, SLIP10Ed25519PrivateKey,
    SLIP10Ed25519Blake2bECC, SLIP10Ed25519Blake2bPoint, SLIP10Ed25519Blake2bPublicKey, SLIP10Ed25519Blake2bPrivateKey,
    SLIP10Ed25519MoneroECC, SLIP10Ed25519MoneroPoint, SLIP10Ed25519MoneroPublicKey, SLIP10Ed25519MoneroPrivateKey,
    SLIP10Nist256p1ECC, SLIP10Nist256p1Point, SLIP10Nist256p1PublicKey, SLIP10Nist256p1PrivateKey,
    SLIP10Secp256k1ECC, SLIP10Secp256k1Point, SLIP10Secp256k1PublicKey, SLIP10Secp256k1PrivateKey
)
from .iecc import (
    IPoint, IPublicKey, IPrivateKey, IEllipticCurveCryptography
)


class ECCS:
    """
    A class that manages a dictionary of ecc classes.

    This class provides methods to retrieve names and classes of various entropy implementations,
    as well as methods to validate and access specific ecc classes by name.

    Here are available ecc names and classes:

    +--------------------------+--------------------------------------------------------------------------+
    | Name                     | Class                                                                    |
    +==========================+==========================================================================+
    | KholawEd25519ECC         | :class:`hdwallet.eccs.kholaw.ed25519.KholawEd25519ECC`                   |
    +--------------------------+--------------------------------------------------------------------------+
    | SLIP10Ed25519ECC         | :class:`hdwallet.eccs.slip10.ed25519.SLIP10Ed25519ECC`                   |
    +--------------------------+--------------------------------------------------------------------------+
    | SLIP10Ed25519Blake2bECC  | :class:`hdwallet.eccs.slip10.ed25519.blake2b.SLIP10Ed25519Blake2bECC`    |
    +--------------------------+--------------------------------------------------------------------------+
    | SLIP10Ed25519MoneroECC   | :class:`hdwallet.eccs.slip10.ed25519.monero.SLIP10Ed25519MoneroECC`      |
    +--------------------------+--------------------------------------------------------------------------+
    | SLIP10Nist256p1ECC       | :class:`hdwallet.eccs.slip10.nist256p1.SLIP10Nist256p1ECC`               |
    +--------------------------+--------------------------------------------------------------------------+
    | SLIP10Secp256k1ECC       | :class:`hdwallet.eccs.slip10.secp256k1.SLIP10Secp256k1ECC`               |
    +--------------------------+--------------------------------------------------------------------------+
    """

    dictionary: Dict[str, Type[IEllipticCurveCryptography]] = {
        KholawEd25519ECC.NAME: KholawEd25519ECC,
        SLIP10Ed25519ECC.NAME: SLIP10Ed25519ECC,
        SLIP10Ed25519Blake2bECC.NAME: SLIP10Ed25519Blake2bECC,
        SLIP10Ed25519MoneroECC.NAME: SLIP10Ed25519MoneroECC,
        SLIP10Nist256p1ECC.NAME: SLIP10Nist256p1ECC,
        SLIP10Secp256k1ECC.NAME: SLIP10Secp256k1ECC
    }

    @classmethod
    def names(cls) -> List[str]:
        """
        Get the names from the class's dictionary.

        :return: A list of names stored as keys in the `dictionary`.
        :rtype: List[str]
        """

        return list(cls.dictionary.keys())

    @classmethod
    def classes(cls) -> List[Type[IEllipticCurveCryptography]]:
        """
        Get the list of elliptic curve cryptography (ECC) classes from the class's dictionary.

        :return: A list of ECC classes stored as values in the `dictionary`.
        :rtype: List[Type[IEllipticCurveCryptography]]
        """

        return list(cls.dictionary.values())

    @classmethod
    def ecc(cls, name: str) -> Type[IEllipticCurveCryptography]:
        """
        Retrieve an elliptic curve cryptography (ECC) class by name.

        :param name: The name of the ECC class to retrieve.
        :type name: str

        :return: The ECC class corresponding to the given name.
        :rtype: Type[IEllipticCurveCryptography]
        """

        if not cls.is_ecc(name=name):
            raise ECCError(
                "Invalid ECC name", expected=cls.names(), got=name
            )

        return cls.dictionary[name]

    @classmethod
    def is_ecc(cls, name: str) -> bool:
        """
        Check if the given name is a valid ECC class name.

        :param name: The name to check.
        :type name: str

        :return: True if the name is a valid ECC class name, False otherwise.
        :rtype: bool
        """

        return name in cls.names()


def validate_and_get_public_key(
    public_key: Union[bytes, str, IPublicKey], public_key_cls: Type[IPublicKey]
) -> IPublicKey:
    """
    Validate and convert the input to an IPublicKey instance.

    :param public_key: The public key to validate and convert. It can be of type bytes, str, or IPublicKey.
    :type public_key: Union[bytes, str, IPublicKey]
    :param public_key_cls: The class to use for creating an IPublicKey instance from bytes.
    :type public_key_cls: Type[IPublicKey]

    :return: A valid IPublicKey instance.
    :rtype: IPublicKey
    """
    
    try:
        if isinstance(public_key, bytes):
            public_key: IPublicKey = public_key_cls.from_bytes(public_key)
        elif isinstance(public_key, str):
            public_key: IPublicKey = public_key_cls.from_bytes(get_bytes(public_key))
        elif not isinstance(public_key, public_key_cls):
            ecc: Type[IEllipticCurveCryptography] = ECCS.ecc(
                name=public_key_cls.name()
            )
            raise PublicKeyError(
                f"Invalid {ecc.NAME} public key instance", expected=public_key_cls, got=type(public_key)
            )
        return public_key
    except ValueError as error:
        raise PublicKeyError("Invalid public key data")

__all__: List[str] = [
    "IPoint", "IPublicKey", "IPrivateKey", "IEllipticCurveCryptography",
    "KholawEd25519Point", "KholawEd25519PublicKey", "KholawEd25519PrivateKey",
    "SLIP10Ed25519Point", "SLIP10Ed25519PublicKey", "SLIP10Ed25519PrivateKey",
    "SLIP10Ed25519Blake2bPoint", "SLIP10Ed25519Blake2bPublicKey", "SLIP10Ed25519Blake2bPrivateKey",
    "SLIP10Ed25519MoneroPoint", "SLIP10Ed25519MoneroPublicKey", "SLIP10Ed25519MoneroPrivateKey",
    "SLIP10Nist256p1Point", "SLIP10Nist256p1PublicKey", "SLIP10Nist256p1PrivateKey",
    "SLIP10Secp256k1Point", "SLIP10Secp256k1PublicKey", "SLIP10Secp256k1PrivateKey",
    "ECCS", "validate_and_get_public_key"
] + [
    cls.__name__ for cls in ECCS.classes()
]



#!/usr/bin/env python3

# Copyright © 2020-2024, Meheret Tesfaye Batu <meherett.batu@gmail.com>
# Distributed under the MIT software license, see the accompanying
# file COPYING or https://opensource.org/license/mit

from typing import Tuple

from .bip44 import BIP44Derivation


class BIP84Derivation(BIP44Derivation):  # https://github.com/bitcoin/bips/blob/master/bip-0084.mediawiki
    """
    This class implements the BIP84 standard for hierarchical deterministic wallets.
    BIP84 defines a specific path structure for deriving keys from a master seed.

    .. note::
        This class inherits from the ``BIP44Derivation`` class, thereby ensuring that all functions are accessible.
    """

    _purpose: Tuple[int, bool] = (84, True)

    @classmethod
    def name(cls) -> str:
        """
        Get the name of the derivation class.

        :return: The name of the derivation class.
        :rtype: str
        """

        return "BIP84"



#!/usr/bin/env python3

# Copyright © 2020-2024, Meheret Tesfaye Batu <meherett.batu@gmail.com>
# Distributed under the MIT software license, see the accompanying
# file COPYING or https://opensource.org/license/mit

from typing import (
    List, Dict, Type
)

from .iseed import ISeed
from ..exceptions import SeedError
from .algorand import AlgorandSeed
from .bip39 import BIP39Seed
from .cardano import CardanoSeed
from .electrum import (
    ElectrumV1Seed, ElectrumV2Seed
)
from .monero import MoneroSeed


class SEEDS:
    """
    A class that manages a dictionary of seeds classes.

    This class provides methods to retrieve names and classes of various seeds implementations,
    as well as methods to validate and access specific seeds classes by name.

    Here are available SEED names:

    +--------------+------------------------------------------------------+
    | Name         | Class                                                |
    +==============+======================================================+
    | Algorand     |  :class:`hdwallet.seeds.algorand.AlgorandSeed`       |
    +--------------+------------------------------------------------------+
    | BIP39        |  :class:`hdwallet.seeds.bip39.BIP39Seed`             |
    +--------------+------------------------------------------------------+
    | Cardano      |  :class:`hdwallet.seeds.cardano.CardanoSeed`         |
    +--------------+------------------------------------------------------+
    | Electrum-V1  |  :class:`hdwallet.seeds.electrum.v1.ElectrumV1Seed`  |
    +--------------+------------------------------------------------------+
    | Electrum-V2  |  :class:`hdwallet.seeds.electrum.v2.ElectrumV2Seed`  |
    +--------------+------------------------------------------------------+
    | Monero       |  :class:`hdwallet.seeds.monero.MoneroSeed`           |
    +--------------+------------------------------------------------------+
    """

    dictionary: Dict[str, Type[ISeed]] = {
        AlgorandSeed.name(): AlgorandSeed,
        BIP39Seed.name(): BIP39Seed,
        CardanoSeed.name(): CardanoSeed,
        ElectrumV1Seed.name(): ElectrumV1Seed,
        ElectrumV2Seed.name(): ElectrumV2Seed,
        MoneroSeed.name(): MoneroSeed
    }

    @classmethod
    def names(cls) -> List[str]:
        """
        Get the list of seed class names.

        :return: A list of seed class names.
        :rtype: List[str]
        """

        return list(cls.dictionary.keys())

    @classmethod
    def classes(cls) -> List[Type[ISeed]]:
        """
        Get the list of seed classes.

        :return: A list of seed classes.
        :rtype: List[Type[ISeed]]
        """

        return list(cls.dictionary.values())

    @classmethod
    def seed(cls, name: str) -> Type[ISeed]:
        """
        Retrieve an seed class by its name.

        :param name: The name of the seed class.
        :type name: str

        :return: The seed class corresponding to the given name.
        :rtype: Type[ISeed]
        """

        if not cls.is_seed(name=name):
            raise SeedError(
                "Invalid seed name", expected=cls.names(), got=name
            )

        return cls.dictionary[name]

    @classmethod
    def is_seed(cls, name) -> bool:
        """
        Check if a given name is a valid seed name.

        :param name: The name to check.
        :type name: str

        :return: True if the name is valid, False otherwise.
        :rtype: bool
        """

        return name in cls.names()


__all__: List[str] = [
    "ISeed", "SEEDS"
] + [
    cls.__name__ for cls in SEEDS.classes()
]



#!/usr/bin/env python3

# Copyright © 2020-2024, Meheret Tesfaye Batu <meherett.batu@gmail.com>
# Distributed under the MIT software license, see the accompanying
# file COPYING or https://opensource.org/license/mit

from typing import (
    Union, Tuple
)

from ..utils import (
    normalize_index, normalize_derivation, index_tuple_to_string
)
from .iderivation import IDerivation


class ElectrumDerivation(IDerivation):
    """
    This class implements the Electrum standard for hierarchical deterministic wallets.
    Electrum defines a specific path structure for deriving keys from a master seed.

    .. note::
        This class inherits from the ``IDerivation`` class, thereby ensuring that all functions are accessible.
    """

    _change: Union[Tuple[int, bool], Tuple[int, int, bool]]
    _address: Union[Tuple[int, bool], Tuple[int, int, bool]]

    def __init__(
        self, change: Union[str, int, Tuple[int, int]] = 0, address: Union[str, int, Tuple[int, int]] = 0
    ) -> None:
        """
        Initialize an ElectrumDerivation object with the given change and address indexes.

        :param change: The change index, which can be a string, integer, or tuple.
                       Default is 0.
        :type change: Union[str, int, Tuple[int, int]]
        :param address: The address index, which can be a string, integer, or tuple.
                        Default is 0.
        :type address: Union[str, int, Tuple[int, int]]

        :return: No return
        :rtype: NoneType
        """

        super(ElectrumDerivation, self).__init__()

        self._change = normalize_index(index=change, hardened=False)
        self._address = normalize_index(index=address, hardened=False)
        self._path, self._indexes, self._derivations = normalize_derivation(path=(
            f"m/{index_tuple_to_string(index=self._change)}/"
            f"{index_tuple_to_string(index=self._address)}"
        ))

    @classmethod
    def name(cls) -> str:
        """
        Get the name of the derivation class.

        :return: The name of the derivation class.
        :rtype: str
        """

        return "Electrum"

    def from_change(self, change: Union[str, int, Tuple[int, int]]) -> "ElectrumDerivation":
        """
        Sets the change index and updates the derivation path accordingly.

        :param change: The change index to set. It can be a string, integer, or tuple representing the index.
        :type change: Union[str, int, Tuple[int, int]]

        :return: The instance of ElectrumDerivation with the updated change index.
        :rtype: ElectrumDerivation
        """

        self._change = normalize_index(index=change, hardened=False)
        self._path, self._indexes, self._derivations = normalize_derivation(path=(
            f"m/{index_tuple_to_string(index=self._change)}/"
            f"{index_tuple_to_string(index=self._address)}"
        ))
        return self

    def from_address(self, address: Union[str, int, Tuple[int, int]]) -> "ElectrumDerivation":
        """
        Sets the address index and updates the derivation path accordingly.

        :param address: The address index to set. It can be a string, integer, or tuple representing the index.
        :type address: Union[str, int, Tuple[int, int]]

        :return: The instance of ElectrumDerivation with the updated address index.
        :rtype: ElectrumDerivation
        """

        self._address = normalize_index(index=address, hardened=False)
        self._path, self._indexes, self._derivations = normalize_derivation(path=(
            f"m/{index_tuple_to_string(index=self._change)}/"
            f"{index_tuple_to_string(index=self._address)}"
        ))
        return self

    def clean(self) -> "ElectrumDerivation":
        """
        Resets the derivation to the default state with change and address indices set to 0.

        :return: The instance of ElectrumDerivation after resetting to default state.
        :rtype: ElectrumDerivation
        """

        self._change = normalize_index(index=0, hardened=False)
        self._address = normalize_index(index=0, hardened=False)
        self._path, self._indexes, self._derivations = normalize_derivation(path=(
            f"m/{index_tuple_to_string(index=self._change)}/"
            f"{index_tuple_to_string(index=self._address)}"
        ))
        return self

    def change(self) -> int:
        """
        Retrieves the change index from the internal `_change` attribute.

        :return: The change index.
        :rtype: int
        """

        return (
            self._change[1] if len(self._change) == 3 else self._change[0]
        )

    def address(self) -> int:
        """
        Retrieves the address index from the internal `_address` attribute.

        :return: The address index.
        :rtype: int
        """

        return (
            self._address[1] if len(self._address) == 3 else self._address[0]
        )



#!/usr/bin/env python3

# Copyright © 2020-2024, Meheret Tesfaye Batu <meherett.batu@gmail.com>
# Distributed under the MIT software license, see the accompanying
# file COPYING or https://opensource.org/license/mit

from typing import (
    Tuple, Optional, Union
)

import struct

from .libs.base58 import (
    check_encode, check_decode
)
from .utils import (
    get_bytes, bytes_to_string, bytes_to_integer, integer_to_bytes
)
from .exceptions import ExtendedKeyError


def serialize(
    version: Union[str, bytes, int],
    depth: int,
    parent_fingerprint: Union[str, bytes],
    index: int,
    chain_code: Union[str, bytes],
    key: Union[str, bytes],
    encoded: bool = False
) -> Optional[str]:
    try:
        raw: bytes = (
            integer_to_bytes(version) if isinstance(version, int) else get_bytes(version) +
            bytes(bytearray([depth])) +
            get_bytes(parent_fingerprint) +
            struct.pack(">L", index) +
            get_bytes(chain_code) +
            get_bytes(key)
        )
        return check_encode(raw) if encoded else bytes_to_string(raw)
    except TypeError:
        return None


def deserialize(
    key: str, encoded: bool = True
) -> Tuple[bytes, int, bytes, int, bytes, bytes]:

    decoded_key: bytes = get_bytes(
        check_decode(key) if encoded else key
    )

    version, depth, parent_fingerprint, index, chain_code, key = (
        decoded_key[:4],
        bytes_to_integer(decoded_key[4:5]),
        decoded_key[5:9],
        struct.unpack(">L", decoded_key[9:13])[0],
        decoded_key[13:45],
        decoded_key[45:]
    )

    return (
        version, depth, parent_fingerprint, index, chain_code, key
    )


def is_valid_key(key: str, encoded: bool = True) -> bool:
    try:
        deserialize(key=key, encoded=encoded)
        return True
    except Exception:
        return False


def is_root_key(key: str, encoded: bool = True) -> bool:

    if not is_valid_key(key=key, encoded=encoded):
        raise ExtendedKeyError("Invalid extended(x) key")

    version, depth, parent_fingerprint, index, chain_code, key = deserialize(
        key=key, encoded=encoded
    )
    return (
        depth == 0 and
        parent_fingerprint == (integer_to_bytes(0x00) * 4) and
        index == 0
    )



#!/usr/bin/env python3

# Copyright © 2020-2025, Meheret Tesfaye Batu <meherett.batu@gmail.com>
# Distributed under the MIT software license, see the accompanying
# file COPYING or https://opensource.org/license/mit

from nacl.signing import VerifyKey

from ....iecc import (
    IPublicKey, IPrivateKey
)
from .....libs.ed25519 import (
    scalar_is_valid, point_scalar_mul_base
)
from .. import SLIP10Ed25519PrivateKey
from .public_key import SLIP10Ed25519MoneroPublicKey


class SLIP10Ed25519MoneroPrivateKey(SLIP10Ed25519PrivateKey):

    @staticmethod
    def name() -> str:
        """
        Get the name of the ecc class.

        :return: The name of the ecc class.
        :rtype: str
        """

        return "SLIP10-Ed25519-Monero"

    @classmethod
    def from_bytes(cls, private_key: bytes) -> IPrivateKey:
        """
        Creates an IPrivateKey instance from a byte sequence.

        :param private_key: The byte sequence representing the private key.
        :type private_key: bytes

        :return: An instance of IPrivateKey created from the provided byte sequence.
        :rtype: IPrivateKey
        """

        if not scalar_is_valid(private_key):
            raise ValueError("Invalid private key bytes")
        return super().from_bytes(private_key)

    def public_key(self) -> IPublicKey:
        """
        Derives the public key from the private key.

        :return: The public key derived from the private key.
        :rtype: IPublicKey
        """

        return SLIP10Ed25519MoneroPublicKey(
            VerifyKey(
                point_scalar_mul_base(bytes(self.signing_key))
            )
        )



#!/usr/bin/env python3

# Copyright © 2020-2025, Meheret Tesfaye Batu <meherett.batu@gmail.com>
# Distributed under the MIT software license, see the accompanying
# file COPYING or https://opensource.org/license/mit

from __future__ import annotations

from typing import Any
from abc import (
    ABC, abstractmethod
)


class IPoint(ABC):

    @staticmethod
    @abstractmethod
    def name() -> str:
        """
        Get the name of the ecc class.

        :return: The name of the ecc class.
        :rtype: str
        """

    @classmethod
    @abstractmethod
    def from_bytes(cls, point: bytes) -> "IPoint":
        """
        Create an IPoint instance from a byte representation.

        :param point: The byte sequence representing the point.
        :type point: bytes

        :return: A new IPoint instance initialized from the byte representation.
        :rtype: IPoint
        """

    @classmethod
    @abstractmethod
    def from_coordinates(cls, x: int, y: int) -> "IPoint":
        """
        Create an IPoint instance from x and y coordinates.

        :param x: The x-coordinate.
        :type x: int
        :param y: The y-coordinate.
        :type y: int

        :return: A new IPoint instance initialized with the given coordinates.
        :rtype: IPoint
        """

    @abstractmethod
    def x(self) -> int:
        """
        Get the x-coordinate value of the point.

        :return: The x-coordinate value.
        :rtype: int
        """

    @abstractmethod
    def y(self) -> int:
        """
        Get the y-coordinate value of the point.

        :return: The y-coordinate value.
        :rtype: int
        """

    def raw(self) -> bytes:
        """
        Get the raw byte representation of the object.

        :return: The raw byte representation of the object.
        :rtype: bytes
        """
        return self.raw_encoded()

    @abstractmethod
    def raw_encoded(self) -> bytes:
        """
        Get the encoded raw byte representation of the object.

        :return: The encoded raw byte representation of the object.
        :rtype: bytes
        """

    @abstractmethod
    def raw_decoded(self) -> bytes:
        """
        Get the decoded raw byte representation of the object.

        :return: The decoded raw byte representation of the object.
        :rtype: bytes
        """

    @abstractmethod
    def underlying_object(self) -> Any:
        """
        Get the underlying object represented by this instance.

        :return: The underlying object represented by this instance.
        :rtype: Any
        """

    @abstractmethod
    def __add__(self, point: "IPoint") -> "IPoint":
        """
        Add another point to this point.

        :param point: The point to add to this point.
        :type point: IPoint

        :return: A new instance of IPoint representing the sum of the points.
        :rtype: IPoint
        """

    @abstractmethod
    def __radd__(self, point: "IPoint") -> "IPoint":
        """
        Add this point to another point on the right-hand side.

        :param point: The point to which this point is added.
        :type point: IPoint

        :return: A new instance of IPoint representing the sum of the points.
        :rtype: IPoint
        """

    @abstractmethod
    def __mul__(self, scalar: int) -> 'IPoint':
        """
        Multiply this point by a scalar.

        :param scalar: The scalar integer to multiply this point by.
        :type scalar: int

        :return: A new instance of IPoint representing the result of the multiplication.
        :rtype: IPoint
        """

    @abstractmethod
    def __rmul__(self, scalar: int) -> 'IPoint':
        """
        Multiply a scalar by this point on the right-hand side.

        :param scalar: The scalar integer to multiply by this point.
        :type scalar: int

        :return: A new instance of IPoint representing the result of the multiplication.
        :rtype: IPoint
        """



#!/usr/bin/env python3

# Copyright © 2020-2025, Meheret Tesfaye Batu <meherett.batu@gmail.com>
# Distributed under the MIT software license, see the accompanying
# file COPYING or https://opensource.org/license/mit

from .ipoint import IPoint
from .ipublic_key import IPublicKey
from .iprivate_key import IPrivateKey


class IEllipticCurveCryptography:
    """
    Interface for Elliptic Curve Cryptography (ECC) representation.

    This interface defines the key components and attributes required
    for elliptic curve cryptography operations. It is intended to be
    implemented by specific ECC classes.
    """

    NAME: str
    ORDER: int
    GENERATOR: IPoint
    POINT: IPoint
    PUBLIC_KEY: IPublicKey
    PRIVATE_KEY: IPrivateKey



#!/usr/bin/env python3

# Copyright © 2020-2024, Meheret Tesfaye Batu <meherett.batu@gmail.com>
# Distributed under the MIT software license, see the accompanying
# file COPYING or https://opensource.org/license/mit

from typing import List

from .v1 import ElectrumV1Seed
from .v2 import ElectrumV2Seed


__all__: List[str] = [
    "ElectrumV1Seed",
    "ElectrumV2Seed"
]



#!/usr/bin/env python3

# Copyright © 2020-2025, Meheret Tesfaye Batu <meherett.batu@gmail.com>
# Distributed under the MIT software license, see the accompanying
# file COPYING or https://opensource.org/license/mit

from typing import (
    Optional, Union, Tuple, Type
)

from ..libs.ed25519 import scalar_reduce, int_decode
from ..eccs import (
    SLIP10Ed25519MoneroECC, IPoint, IPublicKey, IPrivateKey, SLIP10Ed25519MoneroPublicKey, SLIP10Ed25519MoneroPrivateKey
)
from ..seeds import ISeed
from ..crypto import kekkak256
from ..cryptocurrencies.icryptocurrency import INetwork
from ..cryptocurrencies import Monero
from ..derivations import (
    IDerivation, MoneroDerivation
)
from ..exceptions import (
    NetworkError, DerivationError, AddressError, PrivateKeyError, PublicKeyError, SeedError
)
from ..utils import (
    get_bytes, bytes_to_string, integer_to_bytes, bytes_to_integer
)
from ..addresses import MoneroAddress
from .ihd import IHD


class MoneroHD(IHD):

    _network: INetwork
    _seed: Optional[bytes] = None
    _private_key: Optional[bytes] = None

    _spend_private_key: Optional[IPrivateKey]
    _view_private_key: Union[IPrivateKey]
    _spend_public_key: Union[IPublicKey]
    _view_public_key: Union[IPublicKey]

    _derivation: MoneroDerivation

    def __init__(self, network: Union[str, Type[INetwork]] = "mainnet", **kwargs) -> None:
        """
        Initializes a new instance of the Monero class.


        :param network: Network identifier or type implementing the INetwork interface. Default is "mainnet".
        :type network: Union[str, Type[INetwork]]
        :param kwargs: Additional optional parameters:
        """

        super(MoneroHD, self).__init__(
            ecc=SLIP10Ed25519MoneroECC, **kwargs
        )

        try:
            if not isinstance(network, str) and issubclass(network, INetwork):
                network = network.NAME
            if not Monero.NETWORKS.is_network(network=network):
                raise NetworkError(
                    f"Wrong {Monero.NAME} network",
                    expected=Monero.NETWORKS.get_networks(),
                    got=network
                )
            self._network = Monero.NETWORKS.get_network(network=network)
        except TypeError:
            raise NetworkError(
                "Invalid network type", expected=[str, INetwork], got=type(network)
            )

        self._derivation = MoneroDerivation(
            minor=kwargs.get("minor", 1), major=kwargs.get("major", 0)
        )

    @classmethod
    def name(cls) -> str:
        """
        Get the name of the monero class.

        :return: The name of the monero class.
        :rtype: str
        """
        return "Monero"

    def __update__(self) -> "MoneroHD":
        """
        Update the MoneroHD instance using the current derivation parameters.

        :return: Updated MoneroHD instance.
        :rtype: MoneroHD
        """

        self.from_derivation(derivation=self._derivation)
        return self

    def from_seed(self, seed: Union[bytes, str, ISeed], **kwargs) -> "MoneroHD":
        """
        Initialize the MoneroHD instance from a seed value.

        :param seed: Seed value to initialize the HD wallet.
        :type seed: Union[bytes, str, ISeed]
        :param kwargs: Additional keyword arguments for customization.

        :return: Initialized MoneroHD instance.
        :rtype: MoneroHD
        """

        try:
            self._seed = get_bytes(
                seed.seed() if isinstance(seed, ISeed) else seed
            )
            spend_private_key: bytes = (
                self._seed if len(self._seed) == SLIP10Ed25519MoneroPrivateKey.length() else kekkak256(self._seed)
            )
            return self.from_spend_private_key(
                spend_private_key=scalar_reduce(spend_private_key)
            )
        except ValueError as error:
            raise SeedError("Invalid seed data")

    def from_private_key(self, private_key: Union[bytes, str, IPrivateKey]) -> "MoneroHD":
        """
        Initialize the MoneroHD instance from a private key.

        :param private_key: Private key value to initialize the HD wallet.
        :type private_key: Union[bytes, str, IPrivateKey]

        :return: Initialized MoneroHD instance.
        :rtype: MoneroHD
        """

        try:
            self._private_key = (
                private_key.raw() if isinstance(private_key, SLIP10Ed25519MoneroPrivateKey) else get_bytes(private_key)
            )
            return self.from_spend_private_key(
                spend_private_key=scalar_reduce(kekkak256(self._private_key))
            )
        except ValueError as error:
            raise PrivateKeyError("Invalid private key data")

    def from_derivation(self, derivation: IDerivation) -> "MoneroHD":
        """
        Initialize the MoneroHD instance from a derivation object.

        :param derivation: Derivation object to initialize the HD wallet.
        :type derivation: IDerivation

        :return: Initialized MoneroHD instance.
        :rtype: MoneroHD
        """

        if not isinstance(derivation, MoneroDerivation):
            raise DerivationError(
                f"Invalid {self.name()} derivation instance", expected=MoneroDerivation, got=type(derivation)
            )

        self._derivation = derivation
        return self

    def update_derivation(self, derivation: IDerivation) -> "MoneroHD":
        """
        Update the current derivation of the MoneroHD instance.

        :param derivation: New derivation object to update the HD wallet's derivation.
        :type derivation: IDerivation

        :return: Updated MoneroHD instance with the new derivation.
        :rtype: MoneroHD
        """
        return self.from_derivation(derivation=derivation)

    def clean_derivation(self) -> "MoneroHD":
        """
        Clean up the current derivation state of the MoneroHD instance.

        :return: Updated MoneroHD instance with cleaned derivation.
        :rtype: MoneroHD
        """
        self._derivation.clean()
        self.__update__()
        return self

    def from_spend_private_key(self, spend_private_key: Union[bytes, str, IPrivateKey]) -> "MoneroHD":
        """
        Initialize the MoneroHD instance from a spend private key.

        :param spend_private_key: Spend private key to initialize from.
        :type spend_private_key: Union[bytes, str, IPrivateKey]

        :return: Updated MoneroHD instance initialized from the spend private key.
        :rtype: MoneroHD
        """

        if isinstance(spend_private_key, (bytes, str)):
            spend_private_key: IPrivateKey = SLIP10Ed25519MoneroPrivateKey.from_bytes(get_bytes(spend_private_key))

        self._spend_private_key = spend_private_key
        self._view_private_key = SLIP10Ed25519MoneroPrivateKey.from_bytes(
            scalar_reduce(kekkak256(spend_private_key.raw()))
        )
        self._spend_public_key = self._spend_private_key.public_key()
        self._view_public_key = self._view_private_key.public_key()
        return self

    def from_watch_only(
        self, view_private_key: Union[bytes, str, IPrivateKey], spend_public_key: Union[bytes, str, IPublicKey]
    ) -> "MoneroHD":
        """
        Initialize the MoneroHD instance from watch-only keys.

        :param view_private_key: View private key or seed to initialize from.
        :type view_private_key: Union[bytes, str, IPrivateKey]
        :param spend_public_key: Spend public key to initialize from.
        :type spend_public_key: Union[bytes, str, IPublicKey]

        :return: Updated MoneroHD instance initialized from the watch-only keys.
        :rtype: MoneroHD
        """

        try:
            if isinstance(view_private_key, (bytes, str)):
                view_private_key: IPrivateKey = SLIP10Ed25519MoneroPrivateKey.from_bytes(get_bytes(view_private_key))
        except ValueError as error:
            raise PrivateKeyError("Invalid view private key data")

        try:
            if isinstance(spend_public_key, (bytes, str)):
                spend_public_key: IPublicKey = SLIP10Ed25519MoneroPublicKey.from_bytes(get_bytes(spend_public_key))
        except ValueError as error:
            raise PublicKeyError("Invalid spend public key data")

        self._spend_private_key = None
        self._view_private_key = view_private_key
        self._spend_public_key = spend_public_key
        self._view_public_key = self._view_private_key.public_key()
        return self

    def drive(self, minor_index: int, major_index: int) -> Tuple[IPublicKey, IPublicKey]:
        """
        Derives sub-address public keys based on given minor and major indices.

        :param minor_index: Minor index for sub-address derivation.
        :type minor_index: int
        :param major_index: Major index for sub-address derivation.
        :type major_index: int

        :return: Tuple containing the derived sub-address spend and view public keys.
        :rtype: Tuple[IPublicKey, IPublicKey]
        """

        maximum_index: int = 2 ** 32 - 1
        if minor_index < 0 or minor_index > maximum_index:
            raise DerivationError(
                f"Invalid minor index range", expected=f"0-{maximum_index}", got=minor_index
            )
        if major_index < 0 or major_index > maximum_index:
            raise DerivationError(
                f"Invalid major index range", expected=f"0-{maximum_index}", got=major_index
            )

        if minor_index == 0 and major_index == 0:
            return self._spend_public_key, self._view_public_key

        m: int = int_decode(scalar_reduce(
            kekkak256(
                b"SubAddr\x00" +
                self._view_private_key.raw() +
                integer_to_bytes(
                    major_index, bytes_num=4, endianness="little"
                ) + integer_to_bytes(
                    minor_index, bytes_num=4, endianness="little"
                )
            )
        ))

        sub_address_spend_public_key: IPoint = (
            self._spend_public_key.point() + (SLIP10Ed25519MoneroECC.GENERATOR * m)
        )
        sub_address_view_public_key: IPoint = (
            sub_address_spend_public_key * bytes_to_integer(
                self._view_private_key.raw(), endianness="little"
            )
        )

        return (
            SLIP10Ed25519MoneroPublicKey.from_point(sub_address_spend_public_key),
            SLIP10Ed25519MoneroPublicKey.from_point(sub_address_view_public_key)
        )

    def seed(self) -> Optional[str]:
        """
        Retrieves the seed used in this MoneroHD instance.

        :return: Seed used in this instance, or None if not set.
        :rtype: Optional[str]
        """
        return bytes_to_string(self._seed) if self._seed else None

    def private_key(self) -> Optional[str]:
        """
        Retrieves the private key used in this MoneroHD instance.

        :return: Private key used in this instance, or None if not set.
        :rtype: Optional[str]
        """

        return bytes_to_string(self._private_key) if self._private_key else None

    def spend_private_key(self) -> Optional[str]:
        """
        Retrieves the spend private key used in this MoneroHD instance.

        :return: Spend private key used in this instance, or None if not set.
        :rtype: Optional[str]
        """
        return bytes_to_string(self._spend_private_key.raw()) if self._spend_private_key else None

    def view_private_key(self) -> str:
        """
        Retrieves the view private key used in this MoneroHD instance.

        :return: View private key used in this instance.
        :rtype: str
        """

        return bytes_to_string(
            self._view_private_key.raw()
        )

    def spend_public_key(self) -> str:
        """
        Retrieves the spend public key used in this MoneroHD instance.

        :return: Spend public key used in this instance.
        :rtype: str
        """

        return bytes_to_string(self._spend_public_key.raw_compressed())

    def view_public_key(self) -> str:
        """
        Retrieves the view public key used in this MoneroHD instance.

        :return: View public key used in this instance.
        :rtype: str
        """

        return bytes_to_string(self._view_public_key.raw_compressed())

    def primary_address(self) -> str:
        return MoneroAddress.encode(
            spend_public_key=self._spend_public_key,
            view_public_key=self._view_public_key,
            network=self._network.NAME,
            address_type=Monero.ADDRESS_TYPES.STANDARD,
            payment_id=None
        )

    def integrated_address(self, payment_id: Union[bytes, str]) -> str:
        """
        Generates the primary Monero address associated with the spend and view public keys.

        :return: Primary Monero address.
        :rtype: str
        """

        return MoneroAddress.encode(
            spend_public_key=self._spend_public_key,
            view_public_key=self._view_public_key,
            network=self._network.NAME,
            address_type=Monero.ADDRESS_TYPES.INTEGRATED,
            payment_id=get_bytes(payment_id)
        )

    def sub_address(self, minor: Optional[int] = None, major: Optional[int] = None) -> str:
        """
        Generates a sub-address associated with the given minor and major indexes or uses the current derivation indexes.

        :param minor: Minor index for sub-address derivation.
        :type minor: Optional[int]

        :param major: Major index for sub-address derivation.
        :type major: Optional[int]

        :return: Generated sub-address.
        :rtype: str
        """

        if minor is None and major is None:
            minor, major = self._derivation.minor(), self._derivation.major()
        elif (minor and major is None) or (minor is None and major):
            raise DerivationError("Both minor and major indexes are required")

        if minor == 0 and major == 0:
            return self.primary_address()

        spend_public_key, view_public_key = self.drive(
            minor_index=minor, major_index=major
        )

        return MoneroAddress.encode(
            spend_public_key=spend_public_key,
            view_public_key=view_public_key,
            network=self._network.NAME,
            address_type="sub-address",
            payment_id=None
        )

    def address(self, address_type: str, **kwargs) -> str:
        """
        Generates a Monero address of the specified type.

        :param address_type: str - Type of Monero address to generate.
        :type address_type: str

        :param kwargs: Additional keyword arguments depending on the address type:

        :return: str - Generated Monero address.
        """

        if address_type == Monero.ADDRESS_TYPES.STANDARD:
            return self.primary_address()
        elif address_type == Monero.ADDRESS_TYPES.INTEGRATED:
            return self.integrated_address(
                payment_id=kwargs.get("payment_id")
            )
        elif address_type == Monero.ADDRESS_TYPES.SUB_ADDRESS:
            return self.sub_address(
                minor=kwargs.get("minor", self._derivation.minor()),
                major=kwargs.get("major", self._derivation.major())
            )
        raise AddressError(
            f"Invalid {self.name()} address type",
            expected=Monero.ADDRESS_TYPES.get_address_types(),
            got=address_type
        )



#!/usr/bin/env python3

# Copyright © 2020-2024, Meheret Tesfaye Batu <meherett.batu@gmail.com>
# Distributed under the MIT software license, see the accompanying
# file COPYING or https://opensource.org/license/mit

from typing import (
    Tuple, Union, Optional, Dict
)

from ..utils import (
    normalize_index, normalize_derivation, index_tuple_to_string
)
from ..exceptions import DerivationError
from ..cryptocurrencies import Bitcoin
from .iderivation import IDerivation


class CHANGES:

    EXTERNAL_CHAIN: str = "external-chain"
    INTERNAL_CHAIN: str = "internal-chain"


class BIP44Derivation(IDerivation):  # https://github.com/bitcoin/bips/blob/master/bip-0044.mediawiki
    """
    This class implements the BIP44 standard for hierarchical deterministic wallets.
    BIP44 defines a specific path structure for deriving keys from a master seed.

    .. note::
        This class inherits from the ``IDerivation`` class, thereby ensuring that all functions are accessible.

    +-----------------------+------------------+
    | Name                  | Value            |
    +=======================+==================+
    | EXTERNAL_CHAIN        | external-chain   |
    +-----------------------+------------------+
    | INTERNAL_CHAIN        | internal-chain   |
    +-----------------------+------------------+
    """

    _purpose: Tuple[int, bool] = (44, True)

    _coin_type: Tuple[int, bool]
    _account: Union[Tuple[int, bool], Tuple[int, int, bool]]
    _change: Tuple[int, bool]
    _address: Union[Tuple[int, bool], Tuple[int, int, bool]]

    def __init__(
        self,
        coin_type: Union[str, int] = Bitcoin.COIN_TYPE,
        account: Union[str, int, Tuple[int, int]] = 0,
        change: Union[str, int] = CHANGES.EXTERNAL_CHAIN,
        address: Union[str, int, Tuple[int, int]] = 0
    ) -> None:
        """
        Initialize a BIP44 derivation path with specified parameters.

        :param coin_type: The BIP44 coin type index or tuple. Defaults to 0.
        :type coin_type: Union[str, int]
        :param account: The BIP44 account index or tuple. Defaults to 0.
        :type account: Union[str, int, Tuple[int, int]]
        :param change: The BIP44 change index. Can be 'external-chain', 0, '0', 1, or '1'. Defaults to 'external-chain'.
        :type change: Union[str, int]
        :param address: The BIP44 address index or tuple. Defaults to 0.
        :type address: Union[str, int, Tuple[int, int]]

        :return: None
        """
        super(BIP44Derivation, self).__init__()

        self._coin_type = normalize_index(index=coin_type, hardened=True)
        self._account = normalize_index(index=account, hardened=True)
        self._change = normalize_index(
            index=self.get_change_value(change=change, name_only=False), hardened=False
        )
        self._address = normalize_index(index=address, hardened=False)
        self._path, self._indexes, self._derivations = normalize_derivation(path=(
            f"m/{index_tuple_to_string(index=self._purpose)}/"
            f"{index_tuple_to_string(index=self._coin_type)}/"
            f"{index_tuple_to_string(index=self._account)}/"
            f"{index_tuple_to_string(index=self._change)}/"
            f"{index_tuple_to_string(index=self._address)}"
        ))

    @classmethod
    def name(cls) -> str:
        """
        Get the name of the derivation class.

        :return: The name of the derivation class.
        :rtype: str
        """

        return "BIP44"

    def get_change_value(self, change: Union[str, int], name_only: bool = False):
        if isinstance(change, (list, tuple)):
            raise DerivationError(
                "Bad change instance", expected ="int | str", got=type(change).__name__
            )
        external_change = [CHANGES.EXTERNAL_CHAIN, 0, '0']
        internal_change = [CHANGES.INTERNAL_CHAIN, 1, '1']
        expected_change = external_change + internal_change
        if change not in expected_change:
            raise DerivationError(
                f"Bad {self.name()} change index", expected=expected_change, got=change
            )
        if change in external_change:
            return CHANGES.EXTERNAL_CHAIN if name_only else 0
        if change in internal_change:
            return CHANGES.INTERNAL_CHAIN if name_only else 1

    def from_coin_type(self, coin_type: Union[str, int]) -> "BIP44Derivation":
        """
        Set the object's `_coin_type` attribute to the specified coin type index or tuple,
        updating `_path`, `_indexes`, and `_derivations` accordingly.

        :param coin_type: The coin type index or tuple to set. Can be a string or integer.
        :type coin_type: Union[str, int]

        :return: The updated `BIP44Derivation` object itself after setting the coin type.
        :rtype: BIP44Derivation
        """
        self._coin_type = normalize_index(index=coin_type, hardened=True)
        self._path, self._indexes, self._derivations = normalize_derivation(path=(
            f"m/{index_tuple_to_string(index=self._purpose)}/"
            f"{index_tuple_to_string(index=self._coin_type)}/"
            f"{index_tuple_to_string(index=self._account)}/"
            f"{index_tuple_to_string(index=self._change)}/"
            f"{index_tuple_to_string(index=self._address)}"
        ))
        return self

    def from_account(self, account: Union[str, int, Tuple[int, int]]) -> "BIP44Derivation":
        """
        Set the object's `_account` attribute to the specified account index or tuple,
        updating `_path`, `_indexes`, and `_derivations` accordingly.

        :param account: The account index or tuple to set. Can be a string, integer, or tuple of two integers.
        :type account: Union[str, int, Tuple[int, int]]

        :return: The updated `BIP44Derivation` object itself after setting the account.
        :rtype: BIP44Derivation
        """

        self._account = normalize_index(index=account, hardened=True)
        self._path, self._indexes, self._derivations = normalize_derivation(path=(
            f"m/{index_tuple_to_string(index=self._purpose)}/"
            f"{index_tuple_to_string(index=self._coin_type)}/"
            f"{index_tuple_to_string(index=self._account)}/"
            f"{index_tuple_to_string(index=self._change)}/"
            f"{index_tuple_to_string(index=self._address)}"
        ))
        return self

    def from_change(self, change: Union[str, int]) -> "BIP44Derivation":
        """
        Set the object's `_change` attribute to the specified change index or key,
        updating `_path`, `_indexes`, and `_derivations` accordingly.

        :param change: The change index or key to set. Can be a string, integer, or one of the predefined keys.
        :type change: Union[str, int]

        :return: The updated `BIP44Derivation` object itself after setting the change.
        :rtype: BIP44Derivation
        """

        self._change = normalize_index(
            index=self.get_change_value(change=change, name_only=False), hardened=False
        )
        self._path, self._indexes, self._derivations = normalize_derivation(path=(
            f"m/{index_tuple_to_string(index=self._purpose)}/"
            f"{index_tuple_to_string(index=self._coin_type)}/"
            f"{index_tuple_to_string(index=self._account)}/"
            f"{index_tuple_to_string(index=self._change)}/"
            f"{index_tuple_to_string(index=self._address)}"
        ))
        return self

    def from_address(self, address: Union[str, int, Tuple[int, int]]) -> "BIP44Derivation":
        """
        Set the object's `_address` attribute to the specified address index or tuple of indexes,
        updating `_path`, `_indexes`, and `_derivations` accordingly.

        :param address: The address index or tuple of indexes to set. Should be non-hardened.
        :type address: Union[str, int, Tuple[int, int]]

        :return: The updated `BIP44Derivation` object itself after setting the address.
        :rtype: BIP44Derivation
        """

        self._address = normalize_index(index=address, hardened=False)
        self._path, self._indexes, self._derivations = normalize_derivation(path=(
            f"m/{index_tuple_to_string(index=self._purpose)}/"
            f"{index_tuple_to_string(index=self._coin_type)}/"
            f"{index_tuple_to_string(index=self._account)}/"
            f"{index_tuple_to_string(index=self._change)}/"
            f"{index_tuple_to_string(index=self._address)}"
        ))
        return self

    def clean(self) -> "BIP44Derivation":
        """
        Reset the object's attributes related to BIP-44 derivation to their initial states or defaults.

        :return: The updated `BIP44Derivation` object itself after cleaning.
        :rtype: BIP44Derivation
        """

        self._account = normalize_index(index=0, hardened=True)
        self._change = normalize_index(
            index=self.get_change_value(change=CHANGES.EXTERNAL_CHAIN, name_only=False), hardened=False
        )
        self._address = normalize_index(index=0, hardened=False)
        self._path, self._indexes, self._derivations = normalize_derivation(path=(
            f"m/{index_tuple_to_string(index=self._purpose)}/"
            f"{index_tuple_to_string(index=self._coin_type)}/"
            f"{index_tuple_to_string(index=self._account)}/"
            f"{index_tuple_to_string(index=self._change)}/"
            f"{index_tuple_to_string(index=self._address)}"
        ))
        return self

    def purpose(self) -> int:
        """
        Retrieve the purpose value from the object's `_purpose` attribute.

        Returns the first element of `_purpose`.

        :return: The purpose value stored in `_purpose`.
        :rtype: int
        """

        return self._purpose[0]

    def coin_type(self) -> int:
        """
        Retrieve the coin type value from the object's `_coin_type` attribute.

        Returns the first element of `_coin_type`.

        :return: The coin type value stored in `_coin_type`.
        :rtype: int
        """

        return self._coin_type[0]

    def account(self) -> int:
        """
        Retrieve the account from the object.

        :return: The account value.
        :rtype: int
        """

        return (
            self._account[1] if len(self._account) == 3 else self._account[0]
        )

    def change(self, name_only: bool = True) -> str:
        """
        Get the change value.

        :param name_only: Return the change name if True, or its index if False.
        :type name_only: bool
        :return: Change name or index.
        :rtype: str
        """
        return self.get_change_value(change=self._change[0], name_only=name_only)

    def address(self) -> int:
        """
        Retrieve the address from the object.

        :return: The address value.
        :rtype: int
        """

        return (
            self._address[1] if len(self._address) == 3 else self._address[0]
        )



#!/usr/bin/env python3

# Copyright © 2020-2025, Meheret Tesfaye Batu <meherett.batu@gmail.com>
# Distributed under the MIT software license, see the accompanying
# file COPYING or https://opensource.org/license/mit

from typing import Union

import secrets
import math
import re

from ...utils import (
    get_bytes, bytes_to_integer, bytes_to_string, integer_to_bytes
)
from ..ientropy import IEntropy


class ELECTRUM_V2_ENTROPY_STRENGTHS:
    """
    Constants representing the entropy strengths for Electrum-V2.
    """

    ONE_HUNDRED_THIRTY_TWO: int = 132
    TWO_HUNDRED_SIXTY_FOUR: int = 264


class ElectrumV2Entropy(IEntropy):
    """
    Improved security with a BIP32-compatible seed and unique wordlist for mnemonic
    generation, providing better entropy and hierarchical deterministic key derivation.

    .. note::
        This class inherits from the ``IEntropy`` class, thereby ensuring that all functions are accessible.

    Here are available ``ELECTRUM_V2_ENTROPY_STRENGTHS``:

    +--------------------------+-------+
    | Name                     | Value |
    +==========================+=======+
    | ONE_HUNDRED_THIRTY_TWO   |  132  |
    +--------------------------+-------+
    | TWO_HUNDRED_SIXTY_FOUR   |  264  |
    +--------------------------+-------+
    """

    strengths = [
        ELECTRUM_V2_ENTROPY_STRENGTHS.ONE_HUNDRED_THIRTY_TWO,
        ELECTRUM_V2_ENTROPY_STRENGTHS.TWO_HUNDRED_SIXTY_FOUR
    ]

    @classmethod
    def name(cls) -> str:
        """
        Get the name of the entropy class.

        :return: The name of the entropy class.
        :rtype: str
        """

        return "Electrum-V2"

    @classmethod
    def generate(cls, strength: int) -> str:
        """
        Generates a new entropy value with the given strength.

        :param strength: The entropy value.
        :type strength: int

        :return: The generated entropy value for Electrum-V2.
        :rtype: str
        """

        return bytes_to_string(
            integer_to_bytes(
                1 << (strength - 1) | secrets.randbits(strength)  # Ensure bit length equals with given strength
            )
        )

    @classmethod
    def is_valid(cls, entropy: str) -> bool:
        """
        Checks if the given entropy is valid.

        :param entropy: Hex string representing entropy
        :type entropy: str

        :return: True if is valid, False otherwise.
        :rtype: bool
        """

        return isinstance(entropy, str) and bool(re.fullmatch(
            r'^[0-9a-fA-F]+$', entropy
        )) and cls.are_entropy_bits_enough(get_bytes(entropy))

    @classmethod
    def is_valid_strength(cls, strength: int) -> bool:
        """
        Check if the provided strength is valid.

        :param strength: The entropy strength to validate.
        :type strength: int

        :return: True if the strength is valid, False otherwise.
        :rtype: bool
        """

        for _strength in cls.strengths:
            if _strength - 11 <= strength <= _strength:
                return True
        return False

    @classmethod
    def are_entropy_bits_enough(cls, entropy: Union[bytes, int]) -> bool:
        """
        Check if the provided entropy has enough bits.

        :param entropy: The entropy value to check.
        :type entropy: Union[bytes, int]

        :return: True if the strength is valid, False otherwise.
        :rtype: bool
        """

        if isinstance(entropy, bytes):
            entropy: int = bytes_to_integer(entropy)

        entropy_strength: int = 0 if entropy <= 0 else math.floor(math.log(entropy, 2))
        return cls.is_valid_strength(entropy_strength)



#!/usr/bin/env python3

# Copyright © 2020-2023, Meheret Tesfaye Batu <meherett.batu@gmail.com>
# Distributed under the MIT software license, see the accompanying
# file COPYING or https://opensource.org/license/mit

from typing import (
    List, Dict, Type
)

from ..exceptions import MnemonicError
from .algorand import (
    AlgorandMnemonic, ALGORAND_MNEMONIC_WORDS, ALGORAND_MNEMONIC_LANGUAGES
)
from .bip39 import (
    BIP39Mnemonic, BIP39_MNEMONIC_WORDS, BIP39_MNEMONIC_LANGUAGES
)
from .electrum import (
    ElectrumV1Mnemonic, ELECTRUM_V1_MNEMONIC_WORDS, ELECTRUM_V1_MNEMONIC_LANGUAGES,
    ElectrumV2Mnemonic, ELECTRUM_V2_MNEMONIC_WORDS, ELECTRUM_V2_MNEMONIC_LANGUAGES, ELECTRUM_V2_MNEMONIC_TYPES
)
from .monero import (
    MoneroMnemonic, MONERO_MNEMONIC_WORDS, MONERO_MNEMONIC_LANGUAGES
)
from .imnemonic import IMnemonic


class MNEMONICS:
    """
    A class that manages a dictionary of mnemonic classes.

    This class provides methods to retrieve names and classes of various mnemonic implementations,
    as well as methods to validate and access specific mnemonic classes by name.

    Here are available mnemonic names and classes:

    +--------------+------------------------------------------------------------------------+
    | Name         | Class                                                                  |
    +==============+========================================================================+
    | Algorand     |  :class:`hdwallet.mnemonics.algorand.mnemonic.AlgorandMnemonic`        |
    +--------------+------------------------------------------------------------------------+
    | BIP39        |  :class:`hdwallet.mnemonics.bip39.mnemonic.BIP39Mnemonic`              |
    +--------------+------------------------------------------------------------------------+
    | Electrum-V1  |  :class:`hdwallet.mnemonics.electrum.v1.mnemonic.ElectrumV1Mnemonic`   |
    +--------------+------------------------------------------------------------------------+
    | Electrum-V2  |  :class:`hdwallet.mnemonics.electrum.v2.mnemonic.ElectrumV2Mnemonic`   |
    +--------------+------------------------------------------------------------------------+
    | Monero       |  :class:`hdwallet.mnemonics.monero.mnemonic.MoneroMnemonic`            |
    +--------------+------------------------------------------------------------------------+
    """

    dictionary: Dict[str, Type[IMnemonic]] = {
        AlgorandMnemonic.name(): AlgorandMnemonic,
        BIP39Mnemonic.name(): BIP39Mnemonic,
        ElectrumV1Mnemonic.name(): ElectrumV1Mnemonic,
        ElectrumV2Mnemonic.name(): ElectrumV2Mnemonic,
        MoneroMnemonic.name(): MoneroMnemonic
    }

    @classmethod
    def names(cls) -> List[str]:
        """
        Get the list of mnemonic class names.

        :return: A list of mnemonic class names.
        :rtype: List[str]
        """

        return list(cls.dictionary.keys())

    @classmethod
    def classes(cls) -> List[Type[IMnemonic]]:
        """
        Get the list of mnemonic classes.

        :return: A list of mnemonic classes.
        :rtype: List[Type[IMnemonic]]
        """

        return list(cls.dictionary.values())

    @classmethod
    def mnemonic(cls, name: str) -> Type[IMnemonic]:
        """
        Retrieve an mnemonic class by its name.

        :param name: The name of the mnemonic class.
        :type name: str

        :return: The mnemonic class corresponding to the given name.
        :rtype: Type[IMnemonic]
        """

        if not cls.is_mnemonic(name=name):
            raise MnemonicError(
                "Invalid mnemonic name", expected=cls.names(), got=name
            )

        return cls.dictionary[name]

    @classmethod
    def is_mnemonic(cls, name) -> bool:
        """
        Check if a given name is a valid mnemonic name.

        :param name: The name to check.
        :type name: str

        :return: True if the name is valid, False otherwise.
        :rtype: bool
        """

        return name in cls.names()


__all__: List[str] = [
    "IMnemonic",
    "ALGORAND_MNEMONIC_WORDS", "ALGORAND_MNEMONIC_LANGUAGES",
    "BIP39_MNEMONIC_WORDS", "BIP39_MNEMONIC_LANGUAGES",
    "ELECTRUM_V1_MNEMONIC_WORDS", "ELECTRUM_V1_MNEMONIC_LANGUAGES",
    "ELECTRUM_V2_MNEMONIC_WORDS", "ELECTRUM_V2_MNEMONIC_LANGUAGES", "ELECTRUM_V2_MNEMONIC_TYPES",
    "MONERO_MNEMONIC_WORDS", "MONERO_MNEMONIC_LANGUAGES",
    "MNEMONICS"
] + [
    cls.__name__ for cls in MNEMONICS.classes()
]



#!/usr/bin/env python3

# Copyright © 2020-2025, Meheret Tesfaye Batu <meherett.batu@gmail.com>
# Distributed under the MIT software license, see the accompanying
# file COPYING or https://opensource.org/license/mit

from typing import (
    Any, Union
)

from ..eccs import IPublicKey
from ..cryptocurrencies import Avalanche
from ..exceptions import AddressError
from .cosmos import CosmosAddress
from .iaddress import IAddress


class AvalancheAddress(IAddress):

    hrp: str = Avalanche.NETWORKS.MAINNET.HRP
    address_types: dict = {
        "p-chain": Avalanche.PARAMS.ADDRESS_TYPES.P_CHAIN,
        "x-chain": Avalanche.PARAMS.ADDRESS_TYPES.X_CHAIN
    }

    @staticmethod
    def name() -> str:
        """
        Returns the name of the cryptocurrency.

        :return: The name of the cryptocurrency.
        :rtype: str
        """

        return "Avalanche"

    @classmethod
    def encode(cls, public_key: Union[bytes, str, IPublicKey], **kwargs: Any) -> str:
        """
        Encodes the given public key into an Avalanche address.

        :param public_key: The public key to be encoded.
        :type public_key: Union[bytes, str, IPublicKey]
        :param kwargs: Additional keyword arguments, including "address_type".
        :type kwargs: Any

        :return: The encoded Avalanche address.
        :rtype: str
        """

        if not kwargs.get("address_type"):
            address_type: str = cls.address_types[Avalanche.DEFAULT_ADDRESS_TYPE]
        else:
            if kwargs.get("address_type") not in Avalanche.ADDRESS_TYPES.get_address_types():
                raise AddressError(
                    f"Invalid {cls.name()} address type",
                    expected=Avalanche.ADDRESS_TYPES.get_address_types(),
                    got=kwargs.get("address_type")
                )
            address_type: str = cls.address_types[kwargs.get("address_type")]

        return address_type + CosmosAddress.encode(
            public_key=public_key, hrp=cls.hrp
        )

    @classmethod
    def decode(cls, address: str, **kwargs: Any) -> str:
        """
        Decodes the given Avalanche address into its corresponding public key.

        :param address: The Avalanche address to be decoded.
        :type address: str
        :param kwargs: Additional keyword arguments, including "address_type".
        :type kwargs: Any

        :return: The decoded public key.
        :rtype: str
        """

        if not kwargs.get("address_type"):
            address_type: str = cls.address_types[Avalanche.DEFAULT_ADDRESS_TYPE]
        else:
            if kwargs.get("address_type") not in Avalanche.ADDRESS_TYPES.get_address_types():
                raise AddressError(
                    f"Invalid {cls.name()} address type",
                    expected=Avalanche.ADDRESS_TYPES.get_address_types(),
                    got=kwargs.get("address_type")
                )
            address_type: str = cls.address_types[kwargs.get("address_type")]

        prefix_got: str = address[:len(address_type)]
        if address_type != prefix_got:
            raise ValueError(f"Invalid prefix (expected: {address_type}, got: {prefix_got})")
        address_no_prefix: str = address[len(address_type):]

        return CosmosAddress.decode(
            address=address_no_prefix, hrp=cls.hrp
        )



#!/usr/bin/env python3

# Copyright © 2020-2025, Meheret Tesfaye Batu <meherett.batu@gmail.com>
# Distributed under the MIT software license, see the accompanying
# file COPYING or https://opensource.org/license/mit

from typing import Any
from ecdsa.ecdsa import curve_256
from ecdsa.ellipticcurve import (
    Point, PointJacobi
)
from ecdsa import keys

from ....consts import SLIP10_SECP256K1_CONST
from ...iecc import IPoint
from ....utils import (
    bytes_to_integer, integer_to_bytes
)


class SLIP10Nist256p1Point(IPoint):

    point: PointJacobi

    def __init__(self, point: PointJacobi) -> None:
        """
        Initialize the point object.

        :param point: The PointJacobi object representing the point.
        :type point: PointJacobi
        """

        self.point = point

    @staticmethod
    def name() -> str:
        """
        Get the name of the ecc class.

        :return: The name of the ecc class.
        :rtype: str
        """

        return "SLIP10-Nist256p1"

    @classmethod
    def from_bytes(cls, point: bytes) -> IPoint:
        """
        Create an instance of SLIP10Nist256p1Point from bytes encoding.

        :param point: The bytes encoding the point coordinates.
        :type point: bytes

        :return: An instance of SLIP10Nist256p1Point representing the decoded point.
        :rtype: SLIP10Nist256p1Point
        """

        try:
            return cls(
                PointJacobi.from_bytes(
                    curve_256, point
                )
            )
        except keys.MalformedPointError as ex:
            raise ValueError("Invalid point key bytes") from ex
        except AttributeError:
            return cls.from_coordinates(
                bytes_to_integer(point[:SLIP10_SECP256K1_CONST.POINT_COORDINATE_BYTE_LENGTH]),
                bytes_to_integer(point[SLIP10_SECP256K1_CONST.POINT_COORDINATE_BYTE_LENGTH:])
            )

    @classmethod
    def from_coordinates(cls, x: int, y: int) -> IPoint:
        """
        Create an instance of SLIP10Nist256p1Point from x and y coordinates.

        :param x: The x-coordinate of the point.
        :type x: int

        :param y: The y-coordinate of the point.
        :type y: int

        :return: An instance of SLIP10Nist256p1Point representing the specified coordinates.
        :rtype: SLIP10Nist256p1Point
        """

        return cls(
            PointJacobi.from_affine(
                Point(curve_256, x, y)
            )
        )

    def underlying_object(self) -> Any:
        """
        Retrieve the underlying elliptic curve point object.

        This method returns the underlying elliptic curve point object
        represented by this instance.

        :return: The underlying elliptic curve point object.
        :rtype: Any
        """

        return self.point

    def x(self) -> int:
        """
        Get the x-coordinate of the elliptic curve point.

        :return: The x-coordinate of the elliptic curve point.
        :rtype: int
        """

        return self.point.x()

    def y(self) -> int:
        """
        Get the y-coordinate of the elliptic curve point.

        :return: The y-coordinate of the elliptic curve point.
        :rtype: int
        """

        return self.point.y()

    def raw_encoded(self) -> bytes:
        """
        Get the encoded bytes representation of the point.

        :return: Encoded bytes representation of the elliptic curve point.
        :rtype: bytes
        """

        try:
            return self.point.to_bytes("compressed")
        except AttributeError:
            x_bytes = integer_to_bytes(self.point.x(), SLIP10_SECP256K1_CONST.POINT_COORDINATE_BYTE_LENGTH)
            if self.point.y() & 1:
                enc_bytes = b"\x03" + x_bytes
            else:
                enc_bytes = b"\x02" + x_bytes
            return enc_bytes

    def raw_decoded(self) -> bytes:
        """
        Get the decoded bytes representation of the point.

        :return: Decoded bytes representation of the elliptic curve point.
        :rtype: bytes
        """

        try:
            return self.point.to_bytes()
        except AttributeError:
            x_bytes = integer_to_bytes(self.point.x(), SLIP10_SECP256K1_CONST.POINT_COORDINATE_BYTE_LENGTH)
            y_bytes = integer_to_bytes(self.point.y(), SLIP10_SECP256K1_CONST.POINT_COORDINATE_BYTE_LENGTH)

            return x_bytes + y_bytes

    def __add__(self, point: IPoint) -> IPoint:
        """
        Perform addition with another point on the elliptic curve.

        :param point: Another point to add to the current point.
        :type point: IPoint

        :return: A new instance of the same class representing the resulting
                 point after addition.
        :rtype: IPoint
        """

        return self.__class__(self.point + point.underlying_object())

    def __radd__(self, point: IPoint) -> IPoint:
        """
        Perform right addition with another point on the elliptic curve.

        :param point: Another point to add to the current point.
        :type point: IPoint

        :return: A new instance of the same class representing the resulting
                 point after addition.
        :rtype: IPoint
        """

        return self + point

    def __mul__(self, scalar: int) -> IPoint:
        """
        Perform scalar multiplication of the point on the curve.

        :param scalar: The scalar integer to multiply the point by.
        :type scalar: int

        :return: A new instance of the same class representing the resulting
                 point after scalar multiplication.
        :rtype: IPoint
        """

        return self.__class__(self.point * scalar)

    def __rmul__(self, scalar: int) -> IPoint:
        """
        Perform scalar multiplication in the reverse order.

        :param scalar: The scalar integer to multiply the point by.
        :type scalar: int

        :return: The resulting point after scalar multiplication.
        :rtype: IPoint
        """

        return self * scalar



#!/usr/bin/env python3

# Copyright © 2020-2024, Meheret Tesfaye Batu <meherett.batu@gmail.com>
# Distributed under the MIT software license, see the accompanying
# file COPYING or https://opensource.org/license/mit

from typing import (
    Dict, List, Union, Optional
)

import unicodedata

from ....entropies import (
    IEntropy, ElectrumV1Entropy, ELECTRUM_V1_ENTROPY_STRENGTHS
)
from ....exceptions import (
    EntropyError, MnemonicError
)
from ....utils import (
    get_bytes, integer_to_bytes, bytes_to_integer, bytes_to_string
)
from ...imnemonic import IMnemonic


class ELECTRUM_V1_MNEMONIC_WORDS:
    """
    Constants for electrum v1 mnemonic words.
    """

    TWELVE: int = 12


class ELECTRUM_V1_MNEMONIC_LANGUAGES:
    """
    Constants for electrum v1 mnemonic languages.
    """

    ENGLISH: str = "english"


class ElectrumV1Mnemonic(IMnemonic):
    """
    Manages the first version of Electrum mnemonics, used for seed phrases
    in the Electrum Bitcoin wallet, providing a simple and efficient way
    to backup keys.

    Here are available ``ELECTRUM_V1_MNEMONIC_WORDS``:

    +-----------------------+----------------------+
    | Name                  | Value                |
    +=======================+======================+
    | TWELVE                | 12                   |
    +-----------------------+----------------------+

    Here are available ``ELECTRUM_V1_MNEMONIC_LANGUAGES``:

    +-----------------------+----------------------+
    | Name                  | Value                |
    +=======================+======================+
    | ENGLISH               | english              |
    +-----------------------+----------------------+
    """

    words_list: List[int] = [
        ELECTRUM_V1_MNEMONIC_WORDS.TWELVE
    ]
    words_to_entropy_strength: Dict[int, int] = {
        ELECTRUM_V1_MNEMONIC_WORDS.TWELVE: ELECTRUM_V1_ENTROPY_STRENGTHS.ONE_HUNDRED_TWENTY_EIGHT
    }
    languages: List[str] = [
        ELECTRUM_V1_MNEMONIC_LANGUAGES.ENGLISH
    ]
    wordlist_path: Dict[str, str] = {
        ELECTRUM_V1_MNEMONIC_LANGUAGES.ENGLISH: "electrum/v1/wordlist/english.txt"
    }
    words_list_number: int = 1626

    @classmethod
    def name(cls) -> str:
        """
        Get the name of the mnemonic class.

        :return: The name of the mnemonic class.
        :rtype: str
        """

        return "Electrum-V1"

    @classmethod
    def from_words(cls, words: int, language: str) -> str:
        """
        Generates a mnemonic phrase from a specified number of words and language.

        This method generates a mnemonic phrase by first validating the number of words against the allowed word list sizes.
        It then generates entropy based on the specified number of words and uses it to create a mnemonic phrase in the
        specified language.

        :param words: The number of words in the mnemonic phrase.
        :type words: int
        :param language: The language for which to generate the mnemonic phrase.
        :type language: str

        :return: The generated mnemonic phrase.
        :rtype: str
        """

        if words not in cls.words_list:
            raise MnemonicError("Invalid mnemonic words number", expected=cls.words_list, got=words)

        return cls.from_entropy(
            entropy=ElectrumV1Entropy.generate(cls.words_to_entropy_strength[words]), language=language
        )

    @classmethod
    def from_entropy(cls, entropy: Union[str, bytes, IEntropy], language: str, **kwargs) -> str:
        """
        Generates a mnemonic phrase from entropy data.

        This method generates a mnemonic phrase from the given entropy data using the specified language's word list.

        :param entropy: The entropy data used to generate the mnemonic phrase, either as a string or bytes.
        :type entropy: Union[str, bytes, IEntropy]
        :param language: The language for which to generate the mnemonic phrase.
        :type language: str

        :return: The generated mnemonic phrase.
        :rtype: str
        """

        if isinstance(entropy, str) or isinstance(entropy, bytes):
            return cls.encode(entropy=entropy, language=language)
        elif isinstance(entropy, ElectrumV1Entropy):
            return cls.encode(entropy=entropy.entropy(), language=language)
        raise EntropyError(
            "Invalid entropy instance", expected=[str, bytes, ElectrumV1Entropy], got=type(entropy)
        )

    @classmethod
    def encode(cls, entropy: Union[str, bytes], language: str) -> str:
        """
        Generates a mnemonic phrase from entropy data.

        This method generates a mnemonic phrase from the given entropy data using the specified language's word list.

        :param entropy: The entropy data used to generate the mnemonic phrase, either as a string or bytes.
        :type entropy: Union[str, bytes]
        :param language: The language for which to generate the mnemonic phrase.
        :type language: str

        :return: The generated mnemonic phrase.
        :rtype: str
        """

        entropy: bytes = get_bytes(entropy)
        if not ElectrumV1Entropy.is_valid_bytes_strength(len(entropy)):
            raise EntropyError(
                "Wrong entropy strength", expected=ElectrumV1Entropy.strengths, got=(len(entropy) * 8)
            )

        mnemonic: List[str] = []
        words_list: List[str] = cls.normalize(cls.get_words_list_by_language(language=language))
        for index in range(len(entropy) // 4):

            chunk: int = bytes_to_integer(
                entropy[index * 4:(index * 4) + 4], endianness="big"
            )
            word_1_index: int = chunk % len(words_list)
            word_2_index: int = (
                ((chunk // len(words_list)) + word_1_index) % len(words_list)
            )
            word_3_index: int = (
                ((chunk // len(words_list) // len(words_list)) + word_2_index) % len(words_list)
            )
            mnemonic += [
                words_list[word_index] for word_index in (word_1_index, word_2_index, word_3_index)
            ]

        return " ".join(cls.normalize(mnemonic))

    @classmethod
    def decode(
        cls, mnemonic: str, words_list: Optional[List[str]] = None, words_list_with_index: Optional[dict] = None
    ) -> str:
        """
        Decodes a mnemonic phrase back into entropy data.

        This method decodes the given mnemonic phrase into its original entropy data, using the provided
        word list and index mapping or fetching them based on the mnemonic's language if not provided.

        :param mnemonic: The mnemonic phrase to decode, either as a space-separated string or a list of words.
        :type mnemonic: str
        :param words_list: Optional list of valid words for the mnemonic phrase, normalized and in the correct order.
                           If not provided, uses `cls.get_words_list_by_language` to fetch the list based on the default language.
        :type words_list: Optional[List[str]], optional
        :param words_list_with_index: Optional dictionary mapping words to their indices for quick lookup.
                                      If not provided, constructs this mapping based on `words_list`.
        :type words_list_with_index: Optional[dict], optional

        :return: The decoded entropy data as a byte string.
        :rtype: str
        """

        words: list = cls.normalize(mnemonic)
        if len(words) not in cls.words_list:
            raise MnemonicError("Invalid mnemonic words count", expected=cls.words_list, got=len(words))

        if not words_list or not words_list_with_index:
            words_list, language = cls.find_language(mnemonic=words)
            words_list_with_index: dict = {
                words_list[i]: i for i in range(len(words_list))
            }

        entropy: bytes = b""
        for index in range(len(words) // 3):
            word_1, word_2, word_3 = words[index * 3:(index * 3) + 3]

            word_1_index: int = words_list_with_index[word_1]
            word_2_index: int = words_list_with_index[word_2] % len(words_list)
            word_3_index: int = words_list_with_index[word_3] % len(words_list)

            chunk: int = (
                word_1_index +
                (len(words_list) * ((word_2_index - word_1_index) % len(words_list))) +
                (len(words_list) * len(words_list) * ((word_3_index - word_2_index) % len(words_list)))
            )
            entropy += integer_to_bytes(chunk, bytes_num=4, endianness="big")

        return bytes_to_string(entropy)

    @classmethod
    def is_valid(
        cls, mnemonic: Union[str, List[str]], words_list: Optional[List[str]] = None, words_list_with_index: Optional[dict] = None
    ) -> bool:
        """
        Checks if the given mnemonic phrase is valid.

        This method decodes the mnemonic phrase and verifies its validity using the specified word lists and index mappings.

        :param mnemonic: The mnemonic phrase to check, either as a space-separated string or a list of words.
        :type mnemonic: Union[str, List[str]]
        :param words_list: Optional list of valid words for the mnemonic phrase, normalized and in the correct order.
                           If not provided, uses `cls.get_words_list_by_language` to fetch the list based on the default language.
        :type words_list: Optional[List[str]], optional
        :param words_list_with_index: Optional dictionary mapping words to their indices for quick lookup.
                                      If not provided, constructs this mapping based on `words_list`.
        :type words_list_with_index: Optional[dict], optional

        :return: True if the mnemonic phrase is valid, False otherwise.
        :rtype: bool
        """

        try:
            cls.decode(mnemonic=mnemonic, words_list=words_list, words_list_with_index=words_list_with_index)
            return True
        except (ValueError, KeyError, MnemonicError):
            return False

    @classmethod
    def normalize(cls, mnemonic: Union[str, List[str]]) -> List[str]:
        """
        Normalizes the given mnemonic by splitting it into a list of words if it is a string.

        :param mnemonic: The mnemonic value, which can be a single string of words or a list of words.
        :type mnemonic: Union[str, List[str]]

        :return: A list of words from the mnemonic.
        :rtype: List[str]
        """

        mnemonic: list = mnemonic.split() if isinstance(mnemonic, str) else mnemonic
        return list(map(lambda _: unicodedata.normalize("NFKD", _.lower()), mnemonic))



#!/usr/bin/env python3

# Copyright © 2020-2024, Meheret Tesfaye Batu <meherett.batu@gmail.com>
# Distributed under the MIT software license, see the accompanying
# file COPYING or https://opensource.org/license/mit

import json
import click
import sys

from ...mnemonics import (
    IMnemonic,
    AlgorandMnemonic, ALGORAND_MNEMONIC_WORDS, ALGORAND_MNEMONIC_LANGUAGES,
    BIP39Mnemonic, BIP39_MNEMONIC_WORDS, BIP39_MNEMONIC_LANGUAGES,
    ElectrumV1Mnemonic, ELECTRUM_V1_MNEMONIC_WORDS, ELECTRUM_V1_MNEMONIC_LANGUAGES,
    ElectrumV2Mnemonic, ELECTRUM_V2_MNEMONIC_WORDS, ELECTRUM_V2_MNEMONIC_LANGUAGES,
    MoneroMnemonic, MONERO_MNEMONIC_WORDS, MONERO_MNEMONIC_LANGUAGES,
    MNEMONICS
)


def generate_mnemonic(**kwargs) -> None:
    try:
        if not MNEMONICS.is_mnemonic(name=kwargs.get("client")):
            click.echo(click.style(
                f"Wrong mnemonic client, (expected={MNEMONICS.names()}, got='{kwargs.get('client')}')"
            ), err=True)
            sys.exit()

        if kwargs.get("language") is None:  # Set default language
            if kwargs.get("client") == AlgorandMnemonic.name():
                language: str = ALGORAND_MNEMONIC_LANGUAGES.ENGLISH
            elif kwargs.get("client") == BIP39Mnemonic.name():
                language: str = BIP39_MNEMONIC_LANGUAGES.ENGLISH
            elif kwargs.get("client") == ElectrumV1Mnemonic.name():
                language: str = ELECTRUM_V1_MNEMONIC_LANGUAGES.ENGLISH
            elif kwargs.get("client") == ElectrumV2Mnemonic.name():
                language: str = ELECTRUM_V2_MNEMONIC_LANGUAGES.ENGLISH
            elif kwargs.get("client") == MoneroMnemonic.name():
                language: str = MONERO_MNEMONIC_LANGUAGES.ENGLISH
        else:
            language: str = kwargs.get("language")

        if kwargs.get("words") is None:  # Set default words
            if kwargs.get("client") == AlgorandMnemonic.name():
                words: int = ALGORAND_MNEMONIC_WORDS.TWENTY_FIVE
            elif kwargs.get("client") == BIP39Mnemonic.name():
                words: int = BIP39_MNEMONIC_WORDS.TWELVE
            elif kwargs.get("client") == ElectrumV1Mnemonic.name():
                words: int = ELECTRUM_V1_MNEMONIC_WORDS.TWELVE
            elif kwargs.get("client") == ElectrumV2Mnemonic.name():
                words: int = ELECTRUM_V2_MNEMONIC_WORDS.TWELVE
            elif kwargs.get("client") == MoneroMnemonic.name():
                words: int = MONERO_MNEMONIC_WORDS.TWELVE
        else:
            words: int = kwargs.get("words")

        if not MNEMONICS.mnemonic(name=kwargs.get("client")).is_valid_language(language=language):
            click.echo(click.style(
                f"Wrong {kwargs.get('client')} mnemonic language, "
                f"(expected={MNEMONICS.mnemonic(name=kwargs.get('client')).languages}, got='{language}')"
            ), err=True)
            sys.exit()

        if not MNEMONICS.mnemonic(name=kwargs.get("client")).is_valid_words(words=words):
            click.echo(click.style(
                f"Wrong {kwargs.get('client')} mnemonic words, "
                f"(expected={MNEMONICS.mnemonic(name=kwargs.get('client')).words}, got='{words}')"
            ), err=True)
            sys.exit()

        if kwargs.get("entropy"):
            if kwargs.get("client") == ElectrumV2Mnemonic.name():
                mnemonic: IMnemonic = ElectrumV2Mnemonic(
                    mnemonic=ElectrumV2Mnemonic.from_entropy(
                        entropy=kwargs.get("entropy"),
                        language=language,
                        mnemonic_type=kwargs.get("mnemonic_type"),
                        max_attempts=kwargs.get("max_attempts")
                    ),
                    mnemonic_type=kwargs.get("mnemonic_type")
                )
            elif kwargs.get("client") == MoneroMnemonic.name():
                mnemonic: IMnemonic = MoneroMnemonic(
                    mnemonic=MoneroMnemonic.from_entropy(
                        entropy=kwargs.get("entropy"),
                        language=language,
                        checksum=kwargs.get("checksum")
                    )
                )
            else:
                mnemonic: IMnemonic = MNEMONICS.mnemonic(name=kwargs.get("client")).__call__(
                    mnemonic=MNEMONICS.mnemonic(name=kwargs.get("client")).from_entropy(
                        entropy=kwargs.get("entropy"), language=language
                    )
                )
        else:
            if kwargs.get("client") == ElectrumV2Mnemonic.name():
                mnemonic: IMnemonic = ElectrumV2Mnemonic(
                    mnemonic=ElectrumV2Mnemonic.from_words(
                        words=words,
                        language=language,
                        mnemonic_type=kwargs.get("mnemonic_type"),
                        max_attempts=kwargs.get("max_attempts")
                    ),
                    mnemonic_type=kwargs.get("mnemonic_type")
                )
            else:
                mnemonic: IMnemonic = MNEMONICS.mnemonic(name=kwargs.get("client")).__call__(
                    mnemonic=MNEMONICS.mnemonic(name=kwargs.get("client")).from_words(
                        words=words, language=language
                    )
                )
        output: dict = {
            "client": mnemonic.name(),
            "mnemonic": mnemonic.mnemonic(),
            "language": mnemonic.language(),
            "words": mnemonic.words()
        }
        if mnemonic.name() == ElectrumV2Mnemonic.name():
            output["mnemonic_type"] = kwargs.get("mnemonic_type")
        click.echo(json.dumps(
            output, indent=kwargs.get("indent", 4), ensure_ascii=kwargs.get("ensure_ascii", False)
        ))

    except Exception as exception:
        click.echo(click.style(f"Error: {str(exception)}"), err=True)
        sys.exit()



#!/usr/bin/env python3

# Copyright © 2020-2025, Meheret Tesfaye Batu <meherett.batu@gmail.com>
# Distributed under the MIT software license, see the accompanying
# file COPYING or https://opensource.org/license/mit

from typing import (
    Any, Union
)

from ..libs.segwit_bech32 import (
    segwit_encode, segwit_decode
)
from ..eccs import (
    IPoint, IPublicKey, SLIP10Secp256k1ECC, SLIP10Secp256k1Point, SLIP10Secp256k1PublicKey, validate_and_get_public_key
)
from ..cryptocurrencies import Bitcoin
from ..crypto import sha256
from ..utils import (
    get_bytes, integer_to_bytes, bytes_to_integer, bytes_to_string
)
from .iaddress import IAddress


class P2TRAddress(IAddress):
    
    hrp: str = Bitcoin.NETWORKS.MAINNET.HRP
    field_size: int = Bitcoin.PARAMS.FIELD_SIZE
    tap_tweak_sha256: bytes = get_bytes(Bitcoin.PARAMS.TAP_TWEAK_SHA256)
    witness_version: int = Bitcoin.NETWORKS.MAINNET.WITNESS_VERSIONS.P2TR

    @staticmethod
    def name() -> str:
        """
        Returns the name of the cryptocurrency, which is "P2TR".

        :return: The name of the cryptocurrency.
        :rtype: str
        """

        return "P2TR"

    @classmethod
    def tagged_hash(cls, tag: Union[bytes, str], data_bytes: bytes) -> bytes:
        """
        Computes a tagged hash using a double SHA256 hash with the given tag and data.

        :param tag: The tag used for hashing, either as bytes or a string.
        :type tag: Union[bytes, str]
        :param data_bytes: The data bytes to be hashed.
        :type data_bytes: bytes

        :return: The double SHA256 tagged hash.
        :rtype: bytes
        """

        tag_hash = sha256(tag) if isinstance(tag, str) else tag
        return sha256(tag_hash + tag_hash + data_bytes)

    @classmethod
    def hash_tap_tweak(cls, pub_key: IPublicKey) -> bytes:
        """
        Computes a hash using a tagged hash with the tap tweak SHA256
        algorithm and the x-coordinate of the public key.

        :param pub_key: The public key for computing the tweak hash.
        :type pub_key: IPublicKey

        :return: The hashed tap tweak.
        :rtype: bytes
        """

        return cls.tagged_hash(cls.tap_tweak_sha256, integer_to_bytes(pub_key.point().x()))

    @classmethod
    def lift_x(cls, pub_key: IPublicKey) -> IPoint:
        """
        Lifts the x-coordinate of a given public key onto the secp256k1
        elliptic curve to compute its corresponding y-coordinate.

        :param pub_key: The public key whose x-coordinate needs to be lifted.
        :type pub_key: IPublicKey

        :return: The elliptic curve point (x, y) corresponding to the lifted x-coordinate.
        :rtype: IPoint
        """

        p = cls.field_size
        x = pub_key.point().x()
        if x >= p:
            raise ValueError("Unable to compute LiftX point")
        c = (pow(x, 3, p) + 7) % p
        y = pow(c, (p + 1) // 4, p)
        if c != pow(y, 2, p):
            raise ValueError("Unable to compute LiftX point")
        return SLIP10Secp256k1Point.from_coordinates(
            x, y if y % 2 == 0 else p - y
        )

    @classmethod
    def tweak_public_key(cls, pub_key: IPublicKey) -> bytes:
        """
        Tweaks a given public key by hashing its tap tweak and adjusting its x-coordinate on the secp256k1 elliptic curve.

        :param pub_key: The public key to be tweaked.
        :type pub_key: IPublicKey

        :return: The tweaked public key bytes.
        :rtype: bytes
        """
        h = cls.hash_tap_tweak(pub_key)
        out_point = cls.lift_x(pub_key) + (bytes_to_integer(h) * SLIP10Secp256k1ECC.GENERATOR)
        return integer_to_bytes(out_point.x())

    @classmethod
    def encode(cls, public_key: Union[bytes, str, IPublicKey], **kwargs: Any) -> str:
        """
        Encodes a public key into a SegWit address format using specified human-readable part (HRP) and witness version.

        :param public_key: The public key to be encoded.
        :type public_key: Union[bytes, str, IPublicKey]
        :param kwargs: Additional keyword arguments.
            - hrp: Human-readable part (optional).
            - witness_version: SegWit witness version (optional).
        :type kwargs: Any

        :return: The encoded SegWit address.
        :rtype: str
        """

        public_key: IPublicKey = validate_and_get_public_key(
            public_key=public_key, public_key_cls=SLIP10Secp256k1PublicKey
        )
        return segwit_encode(
            kwargs.get("hrp", cls.hrp),
            kwargs.get("witness_version", cls.witness_version),
            cls.tweak_public_key(public_key)
        )

    @classmethod
    def decode(cls, address: str, **kwargs: Any) -> str:
        """
        Decodes a SegWit address into its original public key using the specified human-readable part (HRP).

        :param address: The SegWit address to decode.
        :type address: str
        :param kwargs: Additional keyword arguments.
            - hrp: Human-readable part (optional).
        :type kwargs: Any

        :return: The decoded public key.
        :rtype: str
        """

        witness_version, address_decode = segwit_decode(
            kwargs.get("hrp", cls.hrp), address
        )

        expected_length: int = SLIP10Secp256k1PublicKey.compressed_length() - 1
        if len(address_decode) != expected_length:
            raise ValueError(f"Invalid length (expected: {expected_length}, got: {len(address_decode)})")
        if witness_version != cls.witness_version:
            raise ValueError(f"Invalid witness version (expected: {cls.witness_version}, got: {witness_version})")

        return bytes_to_string(address_decode)



#!/usr/bin/env python3

# Copyright © 2020-2025, Meheret Tesfaye Batu <meherett.batu@gmail.com>
# Distributed under the MIT software license, see the accompanying
# file COPYING or https://opensource.org/license/mit

from typing import (
    Any, Union
)

from ..eccs import (
    IPublicKey, SLIP10Secp256k1PublicKey, validate_and_get_public_key
)
from ..cryptocurrencies import Icon
from ..crypto import sha3_256
from ..utils import (
    get_bytes, bytes_to_string
)
from .iaddress import IAddress


class IconAddress(IAddress):

    address_prefix: str = Icon.PARAMS.ADDRESS_PREFIX
    key_hash_length: int = Icon.PARAMS.KEY_HASH_LENGTH

    @staticmethod
    def name() -> str:
        """
        Returns the name of the blockchain.

        :return: The name of the blockchain.
        :rtype: str
        """

        return "Icon"

    @classmethod
    def encode(cls, public_key: Union[bytes, str, IPublicKey], **kwargs: Any) -> str:
        """
        Encodes the given public key into an Icon address.

        :param public_key: The public key to be encoded.
        :type public_key: Union[bytes, str, IPublicKey]
        :param kwargs: Additional keyword arguments.
        :type kwargs: Any

        :return: The encoded Icon address.
        :rtype: str
        """

        public_key: IPublicKey = validate_and_get_public_key(
            public_key=public_key, public_key_cls=SLIP10Secp256k1PublicKey
        )
        public_key_hash: bytes = sha3_256(
            public_key.raw_uncompressed()[1:]
        )[-cls.key_hash_length:]

        return cls.address_prefix + bytes_to_string(public_key_hash)

    @classmethod
    def decode(cls, address: str, **kwargs: Any) -> str:
        """
        Decodes the given Icon address into its corresponding public key hash.

        :param address: The Icon address to be decoded.
        :type address: str
        :param kwargs: Additional keyword arguments.
        :type kwargs: Any

        :return: The decoded public key hash.
        :rtype: str
        """

        prefix_got: str = address[:len(cls.address_prefix)]
        if cls.address_prefix != prefix_got:
            raise ValueError(f"Invalid prefix (expected: {cls.address_prefix}, got: {prefix_got})")
        address_no_prefix: str = address[len(cls.address_prefix):]

        public_key_hash: bytes = get_bytes(address_no_prefix)

        if len(public_key_hash) != cls.key_hash_length:
            raise ValueError(f"Invalid length (expected: {cls.key_hash_length}, got: {len(public_key_hash)})")

        return bytes_to_string(public_key_hash)



#!/usr/bin/env python3

# Copyright © 2020-2024, Meheret Tesfaye Batu <meherett.batu@gmail.com>
# Distributed under the MIT software license, see the accompanying
# file COPYING or https://opensource.org/license/mit

from tabulate import tabulate

import click

from ...entropies import ENTROPIES


def list_strengths():

    for index, entropy in enumerate(ENTROPIES.classes()):

        strengths: list = []
        for strength in entropy.strengths:
            strengths.append([strength])

        click.echo(tabulate(
            strengths,
            [
                f"{entropy.name()} Strengths"
            ],
            stralign="left",
            numalign="left"
        ))
        if index != (len(ENTROPIES.classes()) - 1):
            click.echo("\n")



#!/usr/bin/env python3

# Copyright © 2020-2025, Meheret Tesfaye Batu <meherett.batu@gmail.com>
# Distributed under the MIT software license, see the accompanying
# file COPYING or https://opensource.org/license/mit

from typing import Union

from ..exceptions import MnemonicError
from ..mnemonics import (
    IMnemonic, AlgorandMnemonic
)
from .iseed import ISeed


class AlgorandSeed(ISeed):
    """
    This class generates a root extended private key from a given seed using the
    Algorand standard. The Algorand standard defines a method for generating mnemonic
    phrases and converting them into a binary seed used for hierarchical
    deterministic wallets.

    .. note::
        This class inherits from the ``ISeed`` class, thereby ensuring that all functions are accessible.
    """

    length = 64

    @classmethod
    def name(cls) -> str:
        """
        Get the name of the seeds class.

        :return: The name of the seeds class.
        :rtype: str
        """

        return "Algorand"

    @classmethod
    def from_mnemonic(cls, mnemonic: Union[str, IMnemonic], **kwargs) -> str:
        """
        Converts a mnemonic phrase to its corresponding seed.

        :param mnemonic: The mnemonic phrase to be decoded. Can be a string or an instance of `IMnemonic`.
        :type mnemonic: Union[str, IMnemonic]

        :return: The decoded entropy as a string.
        :rtype: str
        """

        mnemonic = (
            mnemonic.mnemonic() if isinstance(mnemonic, IMnemonic) else mnemonic
        )
        if not AlgorandMnemonic.is_valid(mnemonic=mnemonic):
            raise MnemonicError(f"Invalid {cls.name()} mnemonic words")

        return AlgorandMnemonic.decode(mnemonic=mnemonic)



#!/usr/bin/env python3

# Copyright © 2020-2024, Meheret Tesfaye Batu <meherett.batu@gmail.com>
# Distributed under the MIT software license, see the accompanying
# file COPYING or https://opensource.org/license/mit

from typing import List

from .mnemonic import (
    ElectrumV1Mnemonic, ELECTRUM_V1_MNEMONIC_WORDS, ELECTRUM_V1_MNEMONIC_LANGUAGES
)


__all__: List[str] = [
    "ElectrumV1Mnemonic",
    "ELECTRUM_V1_MNEMONIC_WORDS",
    "ELECTRUM_V1_MNEMONIC_LANGUAGES"
]



#!/usr/bin/env python3

# Copyright © 2020-2024, Meheret Tesfaye Batu <meherett.batu@gmail.com>
# Distributed under the MIT software license, see the accompanying
# file COPYING or https://opensource.org/license/mit

from typing import (
    List, Dict, Type
)

from ..exceptions import AddressError
from .algorand import AlgorandAddress
from .aptos import AptosAddress
from .avalanche import AvalancheAddress
from .cardano import CardanoAddress
from .cosmos import CosmosAddress
from .eos import EOSAddress
from .ergo import ErgoAddress
from .ethereum import EthereumAddress
from .filecoin import FilecoinAddress
from .harmony import HarmonyAddress
from .icon import IconAddress
from .injective import InjectiveAddress
from .monero import MoneroAddress
from .multiversx import MultiversXAddress
from .nano import NanoAddress
from .near import NearAddress
from .neo import NeoAddress
from .okt_chain import OKTChainAddress
from .p2pkh import P2PKHAddress
from .p2sh import P2SHAddress
from .p2tr import P2TRAddress
from .p2wpkh import P2WPKHAddress
from .p2wpkh_in_p2sh import P2WPKHInP2SHAddress
from .p2wsh import P2WSHAddress
from .p2wsh_in_p2sh import P2WSHInP2SHAddress
from .ripple import RippleAddress
from .solana import SolanaAddress
from .stellar import StellarAddress
from .sui import SuiAddress
from .tezos import TezosAddress
from .tron import TronAddress
from .xinfin import XinFinAddress
from .zilliqa import ZilliqaAddress
from .iaddress import IAddress


class ADDRESSES:
    """
    A class that manages a dictionary of address classes.

    This class provides methods to retrieve names and classes of various address implementations,
    as well as methods to validate and access specific address classes by name.

    Here are available address names and classes:

    +-----------------+------------------------------------------------------------------+
    | Name            | Class                                                            |
    +=================+==================================================================+
    | Algorand        | :class:`hdwallet.addresses.algorand.AlgorandAddress`             |
    +-----------------+------------------------------------------------------------------+
    | Aptos           | :class:`hdwallet.addresses.aptos.AptosAddress`                   |
    +-----------------+------------------------------------------------------------------+
    | Avalanche       | :class:`hdwallet.addresses.avalanche.AvalancheAddress`           |
    +-----------------+------------------------------------------------------------------+
    | Cardano         | :class:`hdwallet.addresses.cardano.CardanoAddress`               |
    +-----------------+------------------------------------------------------------------+
    | Cosmos          | :class:`hdwallet.addresses.cosmos.CosmosAddress`                 |
    +-----------------+------------------------------------------------------------------+
    | EOS             | :class:`hdwallet.addresses.eos.EOSAddress`                       |
    +-----------------+------------------------------------------------------------------+
    | Ergo            | :class:`hdwallet.addresses.ergo.ErgoAddress`                     |
    +-----------------+------------------------------------------------------------------+
    | Ethereum        | :class:`hdwallet.addresses.ethereum.EthereumAddress`             |
    +-----------------+------------------------------------------------------------------+
    | Filecoin        | :class:`hdwallet.addresses.filecoin.FilecoinAddress`             |
    +-----------------+------------------------------------------------------------------+
    | Harmony         | :class:`hdwallet.addresses.harmony.HarmonyAddress`               |
    +-----------------+------------------------------------------------------------------+
    | Icon            | :class:`hdwallet.addresses.icon.IconAddress`                     |
    +-----------------+------------------------------------------------------------------+
    | Injective       | :class:`hdwallet.addresses.injective.InjectiveAddress`           |
    +-----------------+------------------------------------------------------------------+
    | Monero          | :class:`hdwallet.addresses.monero.MoneroAddress`                 |
    +-----------------+------------------------------------------------------------------+
    | MultiversX      | :class:`hdwallet.addresses.multiversx.MultiversXAddress`         |
    +-----------------+------------------------------------------------------------------+
    | Nano            | :class:`hdwallet.addresses.nano.NanoAddress`                     |
    +-----------------+------------------------------------------------------------------+
    | Near            | :class:`hdwallet.addresses.near.NearAddress`                     |
    +-----------------+------------------------------------------------------------------+
    | Neo             | :class:`hdwallet.addresses.neo.NeoAddress`                       |
    +-----------------+------------------------------------------------------------------+
    | OKT-Chain       | :class:`hdwallet.addresses.okt_chain.OKTChainAddress`            |
    +-----------------+------------------------------------------------------------------+
    | P2PKH           | :class:`hdwallet.addresses.p2pkh.P2PKHAddress`                   |
    +-----------------+------------------------------------------------------------------+
    | P2SH            | :class:`hdwallet.addresses.p2sh.P2SHAddress`                     |
    +-----------------+------------------------------------------------------------------+
    | P2TR            | :class:`hdwallet.addresses.p2tr.P2TRAddress`                     |
    +-----------------+------------------------------------------------------------------+
    | P2WPKH          | :class:`hdwallet.addresses.p2wpkh.P2WPKHAddress`                 |
    +-----------------+------------------------------------------------------------------+
    | P2WPKH-In-P2SH  | :class:`hdwallet.addresses.p2wpkh_in_p2sh.P2WPKHInP2SHAddress`   |
    +-----------------+------------------------------------------------------------------+
    | P2WSH           | :class:`hdwallet.addresses.p2wsh.P2WSHAddress`                   |
    +-----------------+------------------------------------------------------------------+
    | P2WSH-In-P2SH   | :class:`hdwallet.addresses.p2wsh_in_p2sh.P2WSHInP2SHAddress`     |
    +-----------------+------------------------------------------------------------------+
    | Ripple          | :class:`hdwallet.addresses.ripple.RippleAddress`                 |
    +-----------------+------------------------------------------------------------------+
    | Solana          | :class:`hdwallet.addresses.solana.SolanaAddress`                 |
    +-----------------+------------------------------------------------------------------+
    | Stellar         | :class:`hdwallet.addresses.stellar.StellarAddress`               |
    +-----------------+------------------------------------------------------------------+
    | Sui             | :class:`hdwallet.addresses.sui.SuiAddress`                       |
    +-----------------+------------------------------------------------------------------+
    | Tezos           | :class:`hdwallet.addresses.tezos.TezosAddress`                   |
    +-----------------+------------------------------------------------------------------+
    | Tron            | :class:`hdwallet.addresses.tron.TronAddress`                     |
    +-----------------+------------------------------------------------------------------+
    | XinFin          | :class:`hdwallet.addresses.xinfin.XinFinAddress`                 |
    +-----------------+------------------------------------------------------------------+
    | Zilliqa         | :class:`hdwallet.addresses.zilliqa.ZilliqaAddress`               |
    +-----------------+------------------------------------------------------------------+

    """

    dictionary: Dict[str, Type[IAddress]] = {
        AlgorandAddress.name(): AlgorandAddress,
        AptosAddress.name(): AptosAddress,
        AvalancheAddress.name(): AvalancheAddress,
        CardanoAddress.name(): CardanoAddress,
        CosmosAddress.name(): CosmosAddress,
        EOSAddress.name(): EOSAddress,
        ErgoAddress.name(): ErgoAddress,
        EthereumAddress.name(): EthereumAddress,
        FilecoinAddress.name(): FilecoinAddress,
        HarmonyAddress.name(): HarmonyAddress,
        IconAddress.name(): IconAddress,
        InjectiveAddress.name(): InjectiveAddress,
        MoneroAddress.name(): MoneroAddress,
        MultiversXAddress.name(): MultiversXAddress,
        NanoAddress.name(): NanoAddress,
        NearAddress.name(): NearAddress,
        NeoAddress.name(): NeoAddress,
        OKTChainAddress.name(): OKTChainAddress,
        P2PKHAddress.name(): P2PKHAddress,
        P2SHAddress.name(): P2SHAddress,
        P2TRAddress.name(): P2TRAddress,
        P2WPKHAddress.name(): P2WPKHAddress,
        P2WPKHInP2SHAddress.name(): P2WPKHInP2SHAddress,
        P2WSHAddress.name(): P2WSHAddress,
        P2WSHInP2SHAddress.name(): P2WSHInP2SHAddress,
        RippleAddress.name(): RippleAddress,
        SolanaAddress.name(): SolanaAddress,
        StellarAddress.name(): StellarAddress,
        SuiAddress.name(): SuiAddress,
        TezosAddress.name(): TezosAddress,
        TronAddress.name(): TronAddress,
        XinFinAddress.name(): XinFinAddress,
        ZilliqaAddress.name(): ZilliqaAddress
    }

    @classmethod
    def names(cls) -> List[str]:
        """
        Get the list of address class names.

        :return: A list of address class names.
        :rtype: List[str]
        """

        return list(cls.dictionary.keys())

    @classmethod
    def classes(cls) -> List[Type[IAddress]]:
        """
        Get the list of address classes.

        :return: A list of address classes.
        :rtype: List[Type[IAddress]]
        """

        return list(cls.dictionary.values())

    @classmethod
    def address(cls, name: str) -> Type[IAddress]:
        """
        Retrieve an address class by its name.

        :param name: The name of the address class.
        :type name: str

        :return: The address class corresponding to the given name.
        :rtype: Type[IAddress]
        """

        if not cls.is_address(name=name):
            raise AddressError(
                "Invalid address name", expected=cls.names(), got=name
            )

        return cls.dictionary[name]

    @classmethod
    def is_address(cls, name: str) -> bool:
        """
        Check if a given name is a valid address name.

        :param name: The name to check.
        :type name: str

        :return: True if the name is valid, False otherwise.
        :rtype: bool
        """

        return name in cls.names()


__all__: List[str] = [
    "IAddress", "ADDRESSES"
] + [
    cls.__name__ for cls in ADDRESSES.classes()
]



#!/usr/bin/env python3

# Copyright © 2020-2024, Meheret Tesfaye Batu <meherett.batu@gmail.com>
#             2024, Eyoel Tadesse <eyoel_tadesse@proton.me>
# Distributed under the MIT software license, see the accompanying
# file COPYING or https://opensource.org/license/mit

import pytest

from hdwallet.exceptions import Error


def test_exception():
	class TestError(Error):
		pass

	with pytest.raises(TestError, match="Error"):
		raise TestError(message="Error")

	with pytest.raises(TestError, match="more_detail"):
		raise TestError(message = "Error", detail="more_detail")

	with pytest.raises(TestError, match="value_expect"):
		raise TestError(message = "Error", expected="value_expect")

	with pytest.raises(TestError, match="value_got"):
		raise TestError(message = "Error", expected="value_expect", got="value_got")



#!/usr/bin/env python3

# Copyright © 2020-2024, Meheret Tesfaye Batu <meherett.batu@gmail.com>
# Distributed under the MIT software license, see the accompanying
# file COPYING or https://opensource.org/license/mit

from typing import (
    Dict, List, Union, Optional
)

import unicodedata

from ....entropies import (
    IEntropy, ElectrumV2Entropy, ELECTRUM_V2_ENTROPY_STRENGTHS
)
from ....crypto import hmac_sha512
from ....exceptions import (
    Error, EntropyError, MnemonicError
)
from ....utils import (
    get_bytes, integer_to_bytes, bytes_to_string, bytes_to_integer
)
from ....mnemonics.bip39 import BIP39Mnemonic
from ....mnemonics.electrum.v1 import ElectrumV1Mnemonic
from ...imnemonic import IMnemonic


class ELECTRUM_V2_MNEMONIC_WORDS:
    """
    Constants for electrum v2 mnemonic words.
    """

    TWELVE: int = 12
    TWENTY_FOUR: int = 24


class ELECTRUM_V2_MNEMONIC_LANGUAGES:
    """
    Constants for electrum v2 mnemonic languages.
    """

    CHINESE_SIMPLIFIED: str = "chinese-simplified"
    ENGLISH: str = "english"
    PORTUGUESE: str = "portuguese"
    SPANISH: str = "spanish"


class ELECTRUM_V2_MNEMONIC_TYPES:
    """
    Constants for electrum v2 mnemonic types.
    """

    STANDARD: str = "standard"
    SEGWIT: str = "segwit"
    STANDARD_2FA: str = "standard-2fa"
    SEGWIT_2FA: str = "segwit-2fa"


class ElectrumV2Mnemonic(IMnemonic):
    """
    An update to the ElectrumV1Mnemonic, this class supports the second version
    of Electrum's mnemonic system, offering enhanced features for secure seed
    generation.

    Here are available ``ELECTRUM_V2_MNEMONIC_LANGUAGES``:

    +-----------------------+----------------------+
    | Name                  | Value                |
    +=======================+======================+
    | TWELVE                | 12                   |
    +-----------------------+----------------------+
    | TWENTY_FOUR           | 24                   |
    +-----------------------+----------------------+

    Here are available ``ELECTRUM_V2_MNEMONIC_LANGUAGES``:

    +-----------------------+----------------------+
    | Name                  | Value                |
    +=======================+======================+
    | CHINESE_SIMPLIFIED    | chinese-simplified   |
    +-----------------------+----------------------+
    | ENGLISH               | english              |
    +-----------------------+----------------------+
    | PORTUGUESE            | portuguese           |
    +-----------------------+----------------------+
    | SPANISH               | spanish              |
    +-----------------------+----------------------+

    Here are available ``ELECTRUM_V2_MNEMONIC_TYPES``:

    +-----------------------+----------------------+
    | Name                  | Value                |
    +=======================+======================+
    | STANDARD              | standard             |
    +-----------------------+----------------------+
    | SEGWIT                | segwit               |
    +-----------------------+----------------------+
    | STANDARD_2FA          | standard-2fa         |
    +-----------------------+----------------------+
    | SEGWIT_2FA            | segwit-2fa           |
    +-----------------------+----------------------+
    """

    word_bit_length: int = 11
    words_list: List[int] = [
        ELECTRUM_V2_MNEMONIC_WORDS.TWELVE,
        ELECTRUM_V2_MNEMONIC_WORDS.TWENTY_FOUR
    ]
    words_to_entropy_strength: Dict[int, int] = {
        ELECTRUM_V2_MNEMONIC_WORDS.TWELVE: ELECTRUM_V2_ENTROPY_STRENGTHS.ONE_HUNDRED_THIRTY_TWO,
        ELECTRUM_V2_MNEMONIC_WORDS.TWENTY_FOUR: ELECTRUM_V2_ENTROPY_STRENGTHS.TWO_HUNDRED_SIXTY_FOUR
    }
    languages: List[str] = [
        ELECTRUM_V2_MNEMONIC_LANGUAGES.CHINESE_SIMPLIFIED,
        ELECTRUM_V2_MNEMONIC_LANGUAGES.ENGLISH,
        ELECTRUM_V2_MNEMONIC_LANGUAGES.PORTUGUESE,
        ELECTRUM_V2_MNEMONIC_LANGUAGES.SPANISH
    ]
    wordlist_path: Dict[str, str] = {
        ELECTRUM_V2_MNEMONIC_LANGUAGES.CHINESE_SIMPLIFIED: "electrum/v2/wordlist/chinese_simplified.txt",
        ELECTRUM_V2_MNEMONIC_LANGUAGES.ENGLISH: "electrum/v2/wordlist/english.txt",
        ELECTRUM_V2_MNEMONIC_LANGUAGES.PORTUGUESE: "electrum/v2/wordlist/portuguese.txt",
        ELECTRUM_V2_MNEMONIC_LANGUAGES.SPANISH: "electrum/v2/wordlist/spanish.txt"
    }
    mnemonic_types: Dict[str, str] = {
        ELECTRUM_V2_MNEMONIC_TYPES.STANDARD: "01",
        ELECTRUM_V2_MNEMONIC_TYPES.SEGWIT: "100",
        ELECTRUM_V2_MNEMONIC_TYPES.STANDARD_2FA: "101",
        ELECTRUM_V2_MNEMONIC_TYPES.SEGWIT_2FA: "102"
    }

    @classmethod
    def name(cls) -> str:
        """
        Get the name of the mnemonic class.

        :return: The name of the entropy class.
        :rtype: str
        """
        return "Electrum-V2"

    @classmethod
    def from_words(
        cls,
        words: int,
        language: str,
        mnemonic_type: str = ELECTRUM_V2_MNEMONIC_TYPES.STANDARD,
        max_attempts: int = 10 ** 60
    ) -> str:
        """
        Generates a mnemonic phrase from a specified number of words and language.

        This method generates a mnemonic phrase with a specific number of words in the specified language.
        It ensures the generated mnemonic meets the strength requirements for the given mnemonic type.

        :param words: The number of words in the mnemonic phrase.
        :type words: int
        :param language: The language for which to generate the mnemonic phrase.
        :type language: str
        :param mnemonic_type: The type of mnemonic phrase to generate. Defaults to `STANDARD`.
        :type mnemonic_type: str, optional
        :param max_attempts: The maximum number of attempts to adjust entropy to meet strength requirements.
                             Defaults to 10^60.
        :type max_attempts: int, optional

        :return: The generated mnemonic phrase.
        :rtype: str

        """

        if words not in cls.words_list:
            raise MnemonicError("Invalid mnemonic words number", expected=cls.words_list, got=words)

        return cls.from_entropy(
            entropy=ElectrumV2Entropy.generate(
                cls.words_to_entropy_strength[words]
            ),
            language=language,
            mnemonic_type=mnemonic_type,
            max_attempts=max_attempts
        )

    @classmethod
    def from_entropy(
        cls,
        entropy: Union[str, bytes, IEntropy],
        language: str,
        mnemonic_type: str = ELECTRUM_V2_MNEMONIC_TYPES.STANDARD,
        max_attempts: int = 10 ** 60
    ) -> str:
        """
        Generates a mnemonic phrase from entropy data, ensuring it meets the required strength.

        This method generates a mnemonic phrase from the provided entropy data, validating
        its strength and ensuring it adheres to the specified mnemonic type and language.

        :param entropy: The entropy data used to generate the mnemonic phrase.
                        It can be a string, bytes, or an instance of IEntropy.
        :type entropy: Union[str, bytes, IEntropy]
        :param language: The language for which to generate the mnemonic phrase.
        :type language: str
        :param mnemonic_type: The type of mnemonic phrase to generate. Defaults to `STANDARD`.
        :type mnemonic_type: str, optional
        :param max_attempts: The maximum number of attempts to adjust entropy to meet strength requirements.
                             Defaults to `10^60`.
        :type max_attempts: int, optional

        :return: The generated mnemonic phrase.
        :rtype: str
        """

        if isinstance(entropy, str) or isinstance(entropy, bytes):
            entropy: bytes = get_bytes(entropy)
        elif isinstance(entropy, ElectrumV2Entropy):
            entropy: bytes = get_bytes(entropy.entropy())
        else:
            raise EntropyError(
                "Invalid entropy instance", expected=[str, bytes, ElectrumV2Entropy], got=type(entropy)
            )

        if ElectrumV2Entropy.are_entropy_bits_enough(entropy):

            words_list: List[str] = cls.normalize(cls.get_words_list_by_language(
                language=language, wordlist_path=cls.wordlist_path
            ))
            bip39_words_list: List[str] = cls.normalize(cls.get_words_list_by_language(
                language=language, wordlist_path=BIP39Mnemonic.wordlist_path
            ))
            bip39_words_list_with_index: dict = {
                bip39_words_list[i]: i for i in range(len(bip39_words_list))
            }
            try:
                electrum_v1_words_list: List[str] = cls.normalize(cls.get_words_list_by_language(
                    language=language, wordlist_path=ElectrumV1Mnemonic.wordlist_path
                ))
                electrum_v1_words_list_with_index: dict = {
                    electrum_v1_words_list[i]: i for i in range(len(electrum_v1_words_list))
                }
            except KeyError:
                electrum_v1_words_list: List[str] = [ ]
                electrum_v1_words_list_with_index: dict = { }

            entropy: int = bytes_to_integer(entropy)
            for index in range(max_attempts):
                new_entropy: int = entropy + index
                try:
                    return cls.encode(
                        entropy=integer_to_bytes(new_entropy),
                        language=language,
                        mnemonic_type=mnemonic_type,
                        words_list=words_list,
                        bip39_words_list=bip39_words_list,
                        bip39_words_list_with_index=bip39_words_list_with_index,
                        electrum_v1_words_list=electrum_v1_words_list,
                        electrum_v1_words_list_with_index=electrum_v1_words_list_with_index
                    )
                except EntropyError:
                    continue

        raise Error("Unable to generate a valid mnemonic")

    @classmethod
    def encode(
        cls,
        entropy: Union[str, bytes],
        language: str,
        mnemonic_type: str = ELECTRUM_V2_MNEMONIC_TYPES.STANDARD,
        words_list: Optional[List[str]] = None,
        bip39_words_list: Optional[List[str]] = None,
        bip39_words_list_with_index: Optional[dict] = None,
        electrum_v1_words_list: Optional[List[str]] = None,
        electrum_v1_words_list_with_index: Optional[dict] = None
    ) -> str:
        """
        Generates a mnemonic phrase from entropy data.

        This method generates a mnemonic phrase from provided entropy data, ensuring
        it meets the requirements of the specified mnemonic type and language.

        :param entropy: The entropy data used to generate the mnemonic phrase.
        :type entropy: Union[str, bytes]
        :param language: The language for which to generate the mnemonic phrase.
        :type language: str
        :param mnemonic_type: The type of mnemonic phrase to generate. Defaults to `STANDARD`.
        :type mnemonic_type: str, optional
        :param words_list: Optional list of words to use for generating the mnemonic.
        :type words_list: List[str], optional
        :param bip39_words_list: Optional list of BIP39 words for validation.
        :type bip39_words_list: List[str], optional
        :param bip39_words_list_with_index: Optional dictionary mapping BIP39 words to indices.
        :type bip39_words_list_with_index: dict, optional
        :param electrum_v1_words_list: Optional list of Electrum V1 words for validation.
        :type electrum_v1_words_list: List[str], optional
        :param electrum_v1_words_list_with_index: Optional dictionary mapping Electrum V1 words to indices.
        :type electrum_v1_words_list_with_index: dict, optional

        :return: The generated mnemonic phrase.
        :rtype: str
        """

        entropy: int = bytes_to_integer(get_bytes(entropy))
        if not ElectrumV2Entropy.are_entropy_bits_enough(entropy):
            raise EntropyError("Entropy bit length is not enough for generating a valid mnemonic")

        mnemonic: List[str] = []
        if not words_list:
            words_list = cls.normalize(cls.get_words_list_by_language(language=language))
        while entropy > 0:
            word_index: int = entropy % len(words_list)
            entropy //= len(words_list)
            mnemonic.append(words_list[word_index])

        if not cls.is_valid(
            mnemonic=mnemonic,
            mnemonic_type=mnemonic_type,
            bip39_words_list=bip39_words_list,
            bip39_words_list_with_index=bip39_words_list_with_index,
            electrum_v1_words_list=electrum_v1_words_list,
            electrum_v1_words_list_with_index=electrum_v1_words_list_with_index
        ):
            raise EntropyError("Entropy bytes are not suitable for generating a valid mnemonic")

        return " ".join(cls.normalize(mnemonic))

    @classmethod
    def decode(cls, mnemonic: str, mnemonic_type: str = ELECTRUM_V2_MNEMONIC_TYPES.STANDARD) -> str:
        """
        Decodes a mnemonic phrase into its original entropy value.

        This method decodes the mnemonic phrase by converting it back into its original
        entropy value, validating its type, and ensuring its integrity.

        :param mnemonic: The mnemonic phrase to decode.
        :type mnemonic: str
        :param mnemonic_type: The type of mnemonic to decode. Defaults to `STANDARD`.
        :type mnemonic_type: str, optional

        :return: The decoded entropy as a string.
        :rtype: str
        """

        words: list = cls.normalize(mnemonic)
        if len(words) not in cls.words_list:
            raise MnemonicError("Invalid mnemonic words count", expected=cls.words_list, got=len(words))

        if not cls.is_valid(mnemonic, mnemonic_type=mnemonic_type):
            raise MnemonicError(f"Invalid {mnemonic_type} mnemonic type words")

        words_list, language = cls.find_language(mnemonic=words)
        words_list_with_index: dict = {
            words_list[i]: i for i in range(len(words_list))
        }

        entropy: int = 0
        for word in reversed(words):
            entropy: int = (entropy * len(words_list)) + words_list_with_index[word]

        return bytes_to_string(integer_to_bytes(entropy))

    @classmethod
    def is_valid(
        cls,
        mnemonic: Union[str, List[str]],
        mnemonic_type: str = ELECTRUM_V2_MNEMONIC_TYPES.STANDARD,
        bip39_words_list: Optional[List[str]] = None,
        bip39_words_list_with_index: Optional[dict] = None,
        electrum_v1_words_list: Optional[List[str]] = None,
        electrum_v1_words_list_with_index: Optional[dict] = None
    ) -> bool:
        """
        Checks if the given mnemonic is valid according to the specified mnemonic type.

        This method validates the mnemonic against BIP39 and Electrum V1 mnemonics,
        and then checks if it matches the specified mnemonic type.

        :param mnemonic: The mnemonic phrase to check.
        :type mnemonic: str or List[str]
        :param mnemonic_type: The type of mnemonic to check against. Defaults to `STANDARD`.
        :type mnemonic_type: str, optional
        :param bip39_words_list: Optional list of BIP39 words for validation.
        :type bip39_words_list: List[str], optional
        :param bip39_words_list_with_index: Optional dictionary mapping BIP39 words to indices.
        :type bip39_words_list_with_index: dict, optional
        :param electrum_v1_words_list: Optional list of Electrum V1 words for validation.
        :type electrum_v1_words_list: List[str], optional
        :param electrum_v1_words_list_with_index: Optional dictionary mapping Electrum V1 words to indices.
        :type electrum_v1_words_list_with_index: dict, optional

        :return: True if the mnemonic is valid according to the specified type, False otherwise.
        :rtype: bool
        """

        if BIP39Mnemonic.is_valid(
            mnemonic, words_list=bip39_words_list, words_list_with_index=bip39_words_list_with_index
        ) or ElectrumV1Mnemonic.is_valid(
            mnemonic, words_list=electrum_v1_words_list, words_list_with_index=electrum_v1_words_list_with_index
        ):
            return False
        return cls.is_type(
            mnemonic=mnemonic, mnemonic_type=mnemonic_type
        )

    @classmethod
    def is_type(
        cls, mnemonic: Union[str, List[str]], mnemonic_type: str = ELECTRUM_V2_MNEMONIC_TYPES.STANDARD
    ) -> bool:
        """
        Checks if the given mnemonic matches the specified mnemonic type.

        :param mnemonic: The mnemonic phrase to check.
        :type mnemonic: str or List[str]
        :param mnemonic_type: The type of mnemonic to check against. Defaults to `STANDARD`.
        :type mnemonic_type: str, optional

        :return: True if the mnemonic matches the specified type, False otherwise.
        :rtype: bool
        """
        return bytes_to_string(hmac_sha512(
            b"Seed version", " ".join(cls.normalize(mnemonic))
        )).startswith(
            cls.mnemonic_types[mnemonic_type]
        )

    def mnemonic_type(self) -> str:
        """
        Retrieves the type of the mnemonic.

        :return: The type of the mnemonic.
        :rtype: str
        """

        return self._mnemonic_type

    @classmethod
    def normalize(cls, mnemonic: Union[str, List[str]]) -> List[str]:
        """
        Normalizes the given mnemonic by splitting it into a list of words if it is a string.

        :param mnemonic: The mnemonic value, which can be a single string of words or a list of words.
        :type mnemonic: Union[str, List[str]]

        :return: A list of words from the mnemonic.
        :rtype: List[str]
        """

        mnemonic: list = mnemonic.split() if isinstance(mnemonic, str) else mnemonic
        return list(map(lambda _: unicodedata.normalize("NFKD", _.lower()), mnemonic))



#!/usr/bin/env python3

# Copyright © 2020-2024, Meheret Tesfaye Batu <meherett.batu@gmail.com>
# Distributed under the MIT software license, see the accompanying
# file COPYING or https://opensource.org/license/mit

from ..cryptocurrencies import Ripple
from .p2pkh import P2PKHAddress


class RippleAddress(P2PKHAddress):

    alphabet: str = Ripple.PARAMS.ALPHABET

    @staticmethod
    def name() -> str:
        """
        Return the name of the cryptocurrency, which is "Ripple".

        :return: The name of the cryptocurrency.
        :rtype: str
        """

        return "Ripple"



#!/usr/bin/env python3

from typing import (
    Tuple, Union, Optional, Literal
)
from binascii import unhexlify
from nacl.bindings import (
    crypto_core_ed25519_add,
    crypto_scalarmult_ed25519_noclamp,
    crypto_scalarmult_ed25519_base_noclamp,
    crypto_core_ed25519_scalar_reduce
)

_Q = 2 ** 255 - 19
_L = 2 ** 252 + 27742317777372353535851937790883648493
_G = (
    15112221349535400772501151409588531511454012693041857206046113283949847762202,
    46316835694926478169428394003475163141307993866256225615783033603165251855960
)
_G_DEC_BYTES = unhexlify(
    "1ad5258f602d56c9b2a7259560c72c695cdcd6fd31e2a4c0fe536ecdd3366921"
    "5866666666666666666666666666666666666666666666666666666666666666"
)
_G_ENC_BYTES = unhexlify(
    "5866666666666666666666666666666666666666666666666666666666666666"
)
_COORD_BYTE_LEN = 32


def bytes_to_integer(data: bytes, endianness: Literal["little", "big"] = "big", signed: bool = False) -> int:
    return int.from_bytes(data, byteorder=endianness, signed=signed)


def integer_to_bytes(data: int, bytes_num: Optional[int] = None, endianness: Literal["little", "big"] = "big", signed: bool = False) -> bytes:
    bytes_num = bytes_num or ((data.bit_length() if data > 0 else 1) + 7) // 8
    return data.to_bytes(bytes_num, byteorder=endianness, signed=signed)


def _inv(x: int) -> int:
    return pow(x, _Q - 2, _Q)


_D = -121665 * _inv(121666)
_I = pow(2, (_Q - 1) // 4, _Q)  # noqa: E741


def _x_recover(y: int) -> int:
    xx = (y * y - 1) * _inv(_D * y * y + 1)
    x = pow(xx, (_Q + 3) // 8, _Q)
    if (x * x - xx) % _Q != 0:
        x = (x * _I) % _Q
    if x % 2 != 0:
        x = _Q - x
    return x


def int_decode(int_bytes: bytes) -> int:
    return bytes_to_integer(int_bytes, endianness="little")


def int_encode(int_val: int) -> bytes:
    return integer_to_bytes(int_val, _COORD_BYTE_LEN, endianness="little")


def point_is_decoded_bytes(point_bytes: bytes) -> bool:
    return len(point_bytes) == _COORD_BYTE_LEN * 2


def point_is_encoded_bytes(point_bytes: bytes) -> bool:
    return len(point_bytes) == _COORD_BYTE_LEN


def point_is_valid_bytes(point_bytes: bytes) -> bool:
    return point_is_decoded_bytes(point_bytes) or point_is_encoded_bytes(point_bytes)


def point_bytes_to_coord(point_bytes: bytes) -> Tuple[int, int]:
    if point_is_decoded_bytes(point_bytes):
        return int_decode(point_bytes[:_COORD_BYTE_LEN]), int_decode(point_bytes[_COORD_BYTE_LEN:])
    if point_is_encoded_bytes(point_bytes):
        return point_decode_no_check(point_bytes)
    raise ValueError("Invalid point bytes")


def point_coord_to_bytes(point_coord: Tuple[int, int]) -> bytes:
    return int_encode(point_coord[0]) + int_encode(point_coord[1])


def point_decode_no_check(point_bytes: bytes) -> Tuple[int, int]:
    if not point_is_encoded_bytes(point_bytes):
        raise ValueError("Invalid point bytes")

    point_int = int_decode(point_bytes)

    clamp = (1 << 255) - 1
    y = point_int & clamp
    x = _x_recover(y)
    if bool(x & 1) != bool(point_int & (1 << 255)):
        x = _Q - x

    return x, y


def point_decode(point_bytes: bytes) -> Tuple[int, int]:
    point_coord = point_decode_no_check(point_bytes)
    if not point_is_on_curve(point_coord):
        raise ValueError("Decoded point does not lie on the curve")
    return point_coord


def point_encode(point_coord: Tuple[int, int]) -> bytes:
    point_bytes = point_coord_to_bytes(point_coord)

    y_bytes = bytearray(point_bytes[_COORD_BYTE_LEN:])
    if point_bytes[0] & 1:
        y_bytes[len(y_bytes) - 1] |= 0x80
    return bytes(y_bytes)


def point_is_generator(point: Union[bytes, Tuple[int, int]]) -> bool:
    # Avoid converting to coordinates if bytes to increase speed
    if isinstance(point, bytes):
        if point_is_encoded_bytes(point):
            return point == _G_ENC_BYTES
        if point_is_decoded_bytes(point):
            return point == _G_DEC_BYTES
        raise ValueError("Invalid point bytes")
    return point == _G


def point_is_on_curve(point: Union[bytes, Tuple[int, int]]) -> bool:
    if isinstance(point, bytes):
        point = point_bytes_to_coord(point)

    x = point[0]
    y = point[1]
    return (-x * x + y * y - 1 - _D * x * x * y * y) % _Q == 0


def point_add(point_1: Union[bytes, Tuple[int, int]],
              point_2: Union[bytes, Tuple[int, int]]) -> bytes:
    return crypto_core_ed25519_add(
        point_1 if isinstance(point_1, bytes) else point_encode(point_1),
        point_2 if isinstance(point_2, bytes) else point_encode(point_2)
    )


def point_scalar_mul(scalar: Union[bytes, int],
                     point: Union[bytes, Tuple[int, int]]) -> bytes:
    return crypto_scalarmult_ed25519_noclamp(
        scalar if isinstance(scalar, bytes) else int_encode(scalar),
        point if isinstance(point, bytes) else point_encode(point)
    )


def point_scalar_mul_base(scalar: Union[bytes, int]) -> bytes:
    return crypto_scalarmult_ed25519_base_noclamp(
        scalar if isinstance(scalar, bytes) else int_encode(scalar)
    )


def scalar_reduce(scalar: Union[bytes, int]) -> bytes:
    if isinstance(scalar, int):
        scalar = int_encode(scalar)
    return crypto_core_ed25519_scalar_reduce(
        scalar.ljust(_COORD_BYTE_LEN * 2, b"\x00")
    )


def scalar_is_valid(scalar: Union[bytes, int]) -> bool:
    if isinstance(scalar, bytes):
        scalar = int_decode(scalar)
    return scalar < _L



#!/usr/bin/env python3

# Copyright © 2020-2024, Meheret Tesfaye Batu <meherett.batu@gmail.com>
# Distributed under the MIT software license, see the accompanying
# file COPYING or https://opensource.org/license/mit

from ..ientropy import IEntropy


class ELECTRUM_V1_ENTROPY_STRENGTHS:
    """
    Constants representing the entropy strengths for Electrum-V1.
    """

    ONE_HUNDRED_TWENTY_EIGHT: int = 128


class ElectrumV1Entropy(IEntropy):
    """
    Relied on user input and simple entropy to generate a seed. It did not use a
    standardized mnemonic, leading to less secure key generation.

    .. note::
        This class inherits from the ``IEntropy`` class, thereby ensuring that all functions are accessible.

    Here are available ``ELECTRUM_V1_ENTROPY_STRENGTHS``:

    +--------------------------+-------+
    | Name                     | Value |
    +==========================+=======+
    | ONE_HUNDRED_TWENTY_EIGHT |  128  |
    +--------------------------+-------+
    """

    strengths = [
        ELECTRUM_V1_ENTROPY_STRENGTHS.ONE_HUNDRED_TWENTY_EIGHT
    ]

    @classmethod
    def name(cls) -> str:
        """
        Get the name of the entropy class.

        :return: The name of the entropy class.
        :rtype: str
        """

        return "Electrum-V1"



#!/usr/bin/env python3

# Copyright © 2020-2024, Meheret Tesfaye Batu <meherett.batu@gmail.com>
# Distributed under the MIT software license, see the accompanying
# file COPYING or https://opensource.org/license/mit

from .ientropy import IEntropy


class BIP39_ENTROPY_STRENGTHS:
    """
    Constants representing the entropy strengths for BIP39.
    """
    ONE_HUNDRED_TWENTY_EIGHT: int = 128
    ONE_HUNDRED_SIXTY: int = 160
    ONE_HUNDRED_NINETY_TWO: int = 192
    TWO_HUNDRED_TWENTY_FOUR: int = 224
    TWO_HUNDRED_FIFTY_SIX: int = 256


class BIP39Entropy(IEntropy):
    """
    Converts entropy into a mnemonic phrase (12, 15, 18, 21, or 24 words). This phrase is used to
    derive a seed, which creates deterministic keys for various cryptocurrencies.

    .. note::
        This class inherits from the ``IEntropy`` class, thereby ensuring that all functions are accessible.

    Here are available ``BIP39_ENTROPY_STRENGTHS``:

    +--------------------------+-------+
    | Name                     | Value |
    +==========================+=======+
    | ONE_HUNDRED_TWENTY_EIGHT |  128  |
    +--------------------------+-------+
    | ONE_HUNDRED_SIXTY        |  160  |
    +--------------------------+-------+
    | ONE_HUNDRED_NINETY_TWO   |  192  |
    +--------------------------+-------+
    | TWO_HUNDRED_TWENTY_FOUR  |  224  |
    +--------------------------+-------+
    | TWO_HUNDRED_FIFTY_SIX    |  256  |
    +--------------------------+-------+
    """

    strengths = [
        BIP39_ENTROPY_STRENGTHS.ONE_HUNDRED_TWENTY_EIGHT,
        BIP39_ENTROPY_STRENGTHS.ONE_HUNDRED_SIXTY,
        BIP39_ENTROPY_STRENGTHS.ONE_HUNDRED_NINETY_TWO,
        BIP39_ENTROPY_STRENGTHS.TWO_HUNDRED_TWENTY_FOUR,
        BIP39_ENTROPY_STRENGTHS.TWO_HUNDRED_FIFTY_SIX
    ]

    @classmethod
    def name(cls) -> str:
        """
        Get the name of the entropy class.

        :return: The name of the entropy class.
        :rtype: str
        """

        return "BIP39"



#!/usr/bin/env python3

# Copyright © 2020-2024, Meheret Tesfaye Batu <meherett.batu@gmail.com>
# Distributed under the MIT software license, see the accompanying
# file COPYING or https://opensource.org/license/mit

from typing import List

# Adcoin
ACC: str = "ACC"
# Akash-Network
AKT: str = "AKT"
# Algorand
ALGO: str = "ALGO"
# Anon
ANON: str = "ANON"
# Aptos
APT: str = "APT"
# Arbitrum
ARB: str = "ARB"
# Argoneum
AGM: str = "AGM"
# Artax
XAX: str = "XAX"
# Aryacoin
AYA: str = "AYA"
# Asiacoin
AC: str = "AC"
# Auroracoin
AUR: str = "AUR"
# Avalanche
AVAX: str = "AVAX"
# Avian
AVN: str = "AVN"
# Axe
AXE: str = "AXE"
# Axelar
AXL: str = "AXL"
# Band-Protocol
BAND: str = "BAND"
# Bata
BTA: str = "BTA"
# Beetle Coin
BEET: str = "BEET"
# Bela Coin
BELA: str = "BELA"
# Binance
BNB: str = "BNB"
# Bit Cloud
BTDX: str = "BTDX"
# Bitcoin
BTC: str = "BTC"
# Bitcoin Atom
BCA: str = "BCA"
# Bitcoin-Cash
BCH: str = "BCH"
# Bitcoin-Cash-SLP
SLP: str = "SLP"
# Bitcoin Gold
BTG: str = "BTG"
# Bitcoin-Green
BITG: str = "BITG"
# Bitcoin Plus
XBC: str = "XBC"
# Bitcoin-Private
BTCP: str = "BTCP"
# Bitcoin SV
BSV: str = "BSV"
# BitcoinZ
BTCZ: str = "BTCZ"
# Bitcore
BTX: str = "BTX"
# Bit Send
BSD: str = "BSD"
# Blackcoin
BLK: str = "BLK"
# Blocknode
BND: str = "BND"
# Block Stamp
BST: str = "BST"
# Bolivarcoin
BOLI: str = "BOLI"
# Brit Coin
BRIT: str = "BRIT"
# Canada eCoin
CDN: str = "CDN"
# Cannacoin
CCN: str = "CCN"
# Cardano
ADA: str = "ADA"
# Celo
CELO: str = "CELO"
# Chihuahua
HUA: str = "HUA"
# Clams
CLAM: str = "CLAM"
# Club Coin
CLUB: str = "CLUB"
# Compcoin
CMP: str = "CMP"
# Cosmos
ATOM: str = "ATOM"
# CPU Chain
CPU: str = "CPU"
# Crane Pay
CRP: str = "CRP"
# Crave
CRAVE: str = "CRAVE"
# Dash
DASH: str = "DASH"
# DeepOnion
ONION: str = "ONION"
# Defcoin
DFC: str = "DFC"
# Denarius
DNR: str = "DNR"
# Diamond
DMD: str = "DMD"
# Digi Byte
DGB: str = "DGB"
# Digitalcoin
DGC: str = "DGC"
# Divi
DIVI: str = "DIVI"
# Dogecoin
DOGE: str = "DOGE"
# dYdX
DYDX: str = "DYDX"
# eCash
XEC: str = "XEC"
# E-coin
ECN: str = "ECN"
# EDR Coin
EDRC: str = "EDRC"
# e-Gulden
EFL: str = "EFL"
# Einsteinium
EMC2: str = "EMC2"
# Elastos
ELA: str = "ELA"
# Energi
NRG: str = "NRG"
# EOS
EOS: str = "EOS"
# Ergo
ERG: str = "ERG"
# Ethereum
ETH: str = "ETH"
# Europe Coin
ERC: str = "ERC"
# Evrmore
EVR: str = "EVR"
# Exclusive Coin
EXCL: str = "EXCL"
# Fantom
FTM: str = "FTM"
# Feathercoin
FTC: str = "FTC"
# Fetch.ai
FET: str = "FET"
# Filecoin
FIL: str = "FIL"
# Firo
FIRO: str = "FIRO"
# Firstcoin
FRST: str = "FRST"
# FIX
FIX: str = "FIX"
# Flashcoin
FLASH: str = "FLASH"
# Flux
FLUX: str = "FLUX"
# Foxdcoin
FOXD: str = "FOXD"
# Fuji Coin
FJC: str = "FJC"
# Game Credits
GAME: str = "GAME"
# GCR Coin
GCR: str = "GCR"
# Go Byte
GBX: str = "GBX"
# Gridcoin
GRC: str = "GRC"
# Groestl Coin
GRS: str = "GRS"
# Gulden
NLG: str = "NLG"
# Harmony
ONE: str = "ONE"
# Helleniccoin
HNC: str = "HNC"
# Hempcoin
THC: str = "THC"
# Horizen
ZEN: str = "ZEN"
# Huobi Token
HT: str = "HT"
# Hush
HUSH: str = "HUSH"
# ICON
ICX: str = "ICX"
# Injective
INJ: str = "INJ"
# InsaneCoin
INSN: str = "INSN"
# Internet Of People
IOP: str = "IOP"
# IRISnet
IRIS: str = "IRIS"
# IX Coin
IXC: str = "IXC"
# Jumbucks
JBS: str = "JBS"
# Kava
KAVA: str = "KAVA"
# Kobocoin
KOBO: str = "KOBO"
# Komodo
KMD: str = "KMD"
# Landcoin
LDCN: str = "LDCN"
# LBRY Credits
LBC: str = "LBC"
# Linx
LINX: str = "LINX"
# Litecoin
LTC: str = "LTC"
# Litecoin Cash
LCC: str = "LCC"
# LitecoinZ
LTZ: str = "LTZ"
# Lkrcoin
LKR: str = "LKR"
# Lynx
LYNX: str = "LYNX"
# Mazacoin
MZC: str = "MZC"
# Megacoin
MEC: str = "MEC"
# Metis
METIS: str = "METIS"
# Minexcoin
MNX: str = "MNX"
# Monacoin
MONA: str = "MONA"
# Monero
XMR: str = "XMR"
# Monk
MONK: str = "MONK"
# MultiversX
EGLD: str = "EGLD"
# Myriadcoin
XMY: str = "XMY"
# Namecoin
NMC: str = "NMC"
# Nano
XNO: str = "XNO"
# Navcoin
NAV: str = "NAV"
# Near
NEAR: str = "NEAR"
# Neblio
NEBL: str = "NEBL"
# Neo
NEO: str = "NEO"
# Neoscoin
NEOS: str = "NEOS"
# Neurocoin
NRO: str = "NRO"
# Neutron
NTRN: str = "NTRN"
# New York Coin
NYC: str = "NYC"
# Nine-Chronicles
NCG: str = "NCG"
# NIX
NIX: str = "NIX"
# Novacoin
NVC: str = "NVC"
# NuBits
NBT: str = "NBT"
# NuShares
NSR: str = "NSR"
# OK Cash
OK: str = "OK"
# OKT-Chain
OKT: str = "OKT"
# Omni
OMNI: str = "OMNI"
# Onix
ONX: str = "ONX"
# Ontology
ONT: str = "ONT"
# Optimism
OP: str = "OP"
# Osmosis
OSMO: str = "OSMO"
# Particl
PART: str = "PART"
# Peercoin
PPC: str = "PPC"
# Pesobit
PSB: str = "PSB"
# Phore
PHR: str = "PHR"
# Pi-Network
PI: str = "PI"
# Pinkcoin
PINK: str = "PINK"
# Pivx
PIVX: str = "PIVX"
# Polygon
MATIC: str = "MATIC"
# PoSW Coin
POSW: str = "POSW"
# Potcoin
POT: str = "POT"
# Project Coin
PRJ: str = "PRJ"
# Putincoin
PUT: str = "PUT"
# Qtum
QTUM: str = "QTUM"
# Rapids
RPD: str = "RPD"
# Ravencoin
RVN: str = "RVN"
# Reddcoin
RDD: str = "RDD"
# Ripple
XRP: str = "XRP"
# Ritocoin
RITO: str = "RITO"
# RSK
RBTC: str = "RBTC"
# Rubycoin
RBY: str = "RBY"
# Safecoin
SAFE: str = "SAFE"
# Saluscoin
SLS: str = "SLS"
# Scribe
SCRIBE: str = "SCRIBE"
# Secret
SCRT: str = "SCRT"
# Shadow Cash
SDC: str = "SDC"
# Shentu
CTK: str = "CTK"
# Slimcoin
SLM: str = "SLM"
# Smileycoin
SMLY: str = "SMLY"
# Solana
SOL: str = "SOL"
# Solarcoin
SLR: str = "SLR"
# Stafi
FIS: str = "FIS"
# Stash
STASH: str = "STASH"
# Stellar
XLM: str = "XLM"
# Stratis
STRAT: str = "STRAT"
# Sugarchain
SUGAR: str = "SUGAR"
# Sui
SUI: str = "SUI"
# Syscoin
SYS: str = "SYS"
# Terra
LUNA: str = "LUNA"
# Tezos
XTZ: str = "XTZ"
# Theta
THETA: str = "THETA"
# Thought AI
THT: str = "THT"
# TOA Coin
TOA: str = "TOA"
# Tron
TRX: str = "TRX"
# TWINS
TWINS: str = "TWINS"
# Ultimate Secure Cash
USC: str = "USC"
# Unobtanium
UNO: str = "UNO"
# Vcash
VC: str = "VC"
# VeChain
VET: str = "VET"
# Verge
XVG: str = "XVG"
# Vertcoin
VTC: str = "VTC"
# Viacoin
VIA: str = "VIA"
# Vivo
VIVO: str = "VIVO"
# Voxels
VOX: str = "VOX"
# Virtual Cash
VASH: str = "VASH"
# Wagerr
WGR: str = "WGR"
# Whitecoin
XWC: str = "XWC"
# Wincoin
WC: str = "WC"
# XinFin
XDC: str = "XDC"
# XUEZ
XUEZ: str = "XUEZ"
# Ycash
YEC: str = "YEC"
# Zcash
ZEC: str = "ZEC"
# ZClassic
ZCL: str = "ZCL"
# Zetacoin
ZET: str = "ZET"
# Zilliqa
ZIL: str = "ZIL"
# ZooBC
ZBC: str = "ZBC"

__all__: List[str] = [
    "ACC",
    "AKT",
    "ALGO",
    "ANON",
    "APT",
    "ARB",
    "AGM",
    "XAX",
    "AYA",
    "AC",
    "AUR",
    "AVAX",
    "AVN",
    "AXE",
    "AXL",
    "BAND",
    "BTA",
    "BEET",
    "BELA",
    "BNB",
    "BTDX",
    "BTC",
    "BCA",
    "BCH",
    "SLP",
    "BTG",
    "BITG",
    "XBC",
    "BTCP",
    "BSV",
    "BTCZ",
    "BTX",
    "BSD",
    "BLK",
    "BND",
    "BST",
    "BOLI",
    "BRIT",
    "CDN",
    "CCN",
    "ADA",
    "CELO",
    "HUA",
    "CLAM",
    "CLUB",
    "CMP",
    "ATOM",
    "CPU",
    "CRP",
    "CRAVE",
    "DASH",
    "ONION",
    "DFC",
    "DNR",
    "DMD",
    "DGB",
    "DGC",
    "DIVI",
    "DOGE",
    "DYDX",
    "XEC",
    "ECN",
    "EDRC",
    "EFL",
    "EMC2",
    "ELA",
    "NRG",
    "EOS",
    "ERG",
    "ETH",
    "ERC",
    "EVR",
    "EXCL",
    "FTM",
    "FTC",
    "FET",
    "FIL",
    "FIRO",
    "FRST",
    "FIX",
    "FLASH",
    "FLUX",
    "FOXD",
    "FJC",
    "GAME",
    "GCR",
    "GBX",
    "GRC",
    "GRS",
    "NLG",
    "ONE",
    "HNC",
    "THC",
    "ZEN",
    "HT",
    "HUSH",
    "ICX",
    "INJ",
    "INSN",
    "IOP",
    "IRIS",
    "IXC",
    "JBS",
    "KAVA",
    "KOBO",
    "KMD",
    "LDCN",
    "LBC",
    "LINX",
    "LTC",
    "LCC",
    "LTZ",
    "LKR",
    "LYNX",
    "MZC",
    "MEC",
    "METIS",
    "MNX",
    "MONA",
    "XMR",
    "MONK",
    "EGLD",
    "XMY",
    "NMC",
    "XNO",
    "NAV",
    "NEAR",
    "NEBL",
    "NEO",
    "NEOS",
    "NRO",
    "NTRN",
    "NYC",
    "NCG",
    "NIX",
    "NVC",
    "NBT",
    "NSR",
    "OK",
    "OKT",
    "OMNI",
    "ONX",
    "ONT",
    "OP",
    "OSMO",
    "PART",
    "PPC",
    "PSB",
    "PHR",
    "PI",
    "PINK",
    "PIVX",
    "MATIC",
    "POSW",
    "POT",
    "PRJ",
    "PUT",
    "QTUM",
    "RPD",
    "RVN",
    "RDD",
    "XRP",
    "RITO",
    "RBTC",
    "RBY",
    "SAFE",
    "SLS",
    "SCRIBE",
    "SCRT",
    "SDC",
    "CTK",
    "SLM",
    "SMLY",
    "SOL",
    "SLR",
    "FIS",
    "STASH",
    "XLM",
    "STRAT",
    "SUGAR",
    "SUI",
    "SYS",
    "LUNA",
    "XTZ",
    "THETA",
    "THT",
    "TOA",
    "TRX",
    "TWINS",
    "USC",
    "UNO",
    "VC",
    "VET",
    "XVG",
    "VTC",
    "VIA",
    "VIVO",
    "VOX",
    "VASH",
    "WGR",
    "XWC",
    "WC",
    "XDC",
    "XUEZ",
    "YEC",
    "ZEC",
    "ZCL",
    "ZET",
    "ZIL",
    "ZBC"
]
